from __future__ import annotations

import logging
import math
from datetime import datetime
from typing import TYPE_CHECKING, Callable, Sequence

if TYPE_CHECKING:
    from ..platform import Platform

import numpy as np

from skretrieval.geodetic import geodetic
from skretrieval.time import ut_to_datetime64

# from ..platform import Platform
from ..platform.rotationmatrix import UnitVectors
from .standard_orientation_techniques import (
    _technique_set_boresight_look_at_location_llh,
    _technique_set_boresight_look_at_location_orbitangle,
    _technique_set_boresight_look_at_location_xyz,
    _technique_set_boresight_pointing_from_unitvectors,
    _technique_set_icf_look_from_xyz,
    _technique_set_icf_orientation_from_azi_elev,
    _technique_set_icf_orientation_no_operation,
    _technique_set_limb_look_vectors_from_unit_xyz,
    _technique_set_look_vectors_from_tangent_altitude,
    _technique_set_observer_to_look_in_azi_elev,
    _technique_set_platform_pointing_from_platform,
    _technique_set_platform_position_from_llh,
    _technique_set_platform_position_from_observer_looking_at_llh,
    _technique_set_platform_position_from_platform,
    _technique_set_platform_position_from_xyz,
)


# ------------------------------------------------------------------------------
#           class PointingAlgorithms:
# ------------------------------------------------------------------------------
class PointingAlgorithms:
    def __init__(self, platform: Platform):
        self._geo = geodetic()
        self.platform: Platform = platform

    # ------------------------------------------------------------------------------
    #           geo
    # ------------------------------------------------------------------------------
    @property
    def geo(self):
        return self._geo

    # -----------------------------------------------------------------------------
    #           set_limb_boresight_to_look_at_tangent_altitude
    # -----------------------------------------------------------------------------

    def set_limb_boresight_to_look_at_tangent_altitude(
        self,
        observer: np.ndarray,
        target_tangent_altitude: float,
        target_geographic_bearing: float,
        roll_control: str,
        roll_angle_degrees: float,
    ) -> bool:
        """
        Sets the rotation matrix of the system so the boresight of the instrument is looking at a tangent point
        at the given tangent altitude and bearing. The tangent altitude must be below the observer and above 5 km below
        the ground.

        Parameters
        ----------
        observer: np.ndarray[3,]
            The x,y,z location of the observer in meters in the :ref:`ecef` reference frame.
        target_tangent_altitude : float
            The tangent altitude of the target location in meteres above sea-level. The tangent altitude  must be below the
            observer and above 5 km below the ground otherwise it is not used.
        target_geographic_bearing: float
            The geographic bearing in degrres of the tangent altitude measured the current platform location. N=0, E=90, S=180, W=270.
        roll_control: str
            The string describing roll control, see: :ref:`rollcontrol`
        roll_angle_degrees: float
            The roll angle in diegrees in the clockwise direction from the zero point set by `roll_control`

        Returns
        -------
        bool: True if successful.
        """

        self._geo.from_xyz(observer)
        north = -self._geo.local_south
        east = -self._geo.local_west
        maxaltitude = self._geo.altitude
        ok = (target_tangent_altitude < maxaltitude) and (
            target_tangent_altitude > -5000.0
        )  # make sure the target tangent altitude is below the observer and no more than 5 km below the surface.
        if ok:
            bearing = math.radians(target_geographic_bearing)
            horiz = math.cos(bearing) * north + math.sin(bearing) * east
            look = self._geo.from_tangent_altitude(
                target_tangent_altitude, observer, horiz
            )
            target = self._geo.location
            G = self.apply_roll_control(
                look,
                target,
                observer,
                roll_control,
                roll_angle_degrees,
                horizontalunitvector=horiz,
            )
            self.platform.platform_pointing.force_pcf_rotation_matrix(G)
        else:
            logging.warning(
                "PointingAlgorithms.set_limb_boresight_to_look_at_tangent_altitude, the target altitude {} is not between the observer altitude {} and 5 km below the ground".format(  # noqa: G001
                    target_tangent_altitude, maxaltitude
                )
            )
        return ok

    # -----------------------------------------------------------------------------
    #           set_limb_boresight_from_lookvector
    # -----------------------------------------------------------------------------

    def set_limb_boresight_from_lookvector(
        self,
        observer: np.ndarray,
        lookvector: np.ndarray,
        roll_control: str,
        roll_angle_degrees: float,
    ) -> bool:
        """
        Sets the platform rotation matrices so the boresight of the instrument is looking in the direction given by lookvector. The
        tangent point, implied by the look vector, must be in front of the observer and above 5 km below the ground.

        Parameters
        ----------
        observer: np.ndarray[3,]
            The x,y,z location of the observer in meters in the :ref:`ecef` reference frame.
        lookvector: np.ndarray[3]
            The x,y,z look unit vector expressed in the :ref:`ecef` reference frame, looking from the platform towards the limb
        roll_control: str
            The string describing roll control, see: :ref:`rollcontrol`
        roll_angle_degrees: float
            The roll angle in diegrees in the clockwise direction from the zero point set by `roll_control`

        Returns
        -------
        bool: True if successful.
        """
        self._geo.from_tangent_point(observer, lookvector)
        targetlocation = self._geo.location
        altitude = self._geo.altitude
        delta = targetlocation - observer
        ok = (np.dot(delta, lookvector) > 0.0) and (altitude > -5000.0)
        if ok:
            G = self.apply_roll_control(
                lookvector, targetlocation, observer, roll_control, roll_angle_degrees
            )
            self.platform.platform_pointing.force_pcf_rotation_matrix(G)
        else:
            logging.warning(
                "PointingAlgorithms.set_limb_boresight_from_lookvector, the tangent point is not in front of the observer or is more than 5000m below the ground [{}]. It has been discarded".format(  # noqa: G001
                    altitude
                )
            )
        return ok

    # ------------------------------------------------------------------------------
    #           get_unitvectors_at_target_location
    # ------------------------------------------------------------------------------

    def set_boresight_to_look_at_geocentric_location(
        self,
        observer_location: np.ndarray,
        targetlocation: np.ndarray,
        roll_control: str,
        roll_angle_degrees: float,
    ) -> bool:
        """
        Sets the rotation matrix of the platform so the boresight of the instrument will look at the given geocentric
        location. The target location must be more than 10 cm from the observer's location.

        Parameters
        ----------
        observer_location: np.ndarray[3,]
            The x,y,z location of the observer in meters in the :ref:`ecef` reference frame.
        targetlocation: np.ndarray[3,]
            The x,y,z location of the target point in meters in the :ref:`ecef` reference frame
        roll_control: str
            The string describing roll control, see: :ref:`rollcontrol`
        roll_angle_degrees: float
            The roll angle in diegrees in the clockwise direction from the zero point set by `roll_control`

        Returns
        -------
        bool: True if successful.
        """
        lookvector = (
            targetlocation - observer_location
        )  # Get the look vector from current location to the target location
        dist = np.linalg.norm(
            lookvector
        )  # Get the distance of the target from the observer
        ok = dist > 0.01  # make sure its more than 10 cm
        if ok:  # and if we are good
            lookvector /= dist  # the normalize the lookvector
            G = self.apply_roll_control(
                lookvector,
                targetlocation,
                observer_location,
                roll_control,
                roll_angle_degrees,
            )  # get the the desired unit vectors
            self.platform.platform_pointing.force_pcf_rotation_matrix(
                G
            )  # and calculate the new rotation matrix from the platform control frame.
        else:
            logging.warning(
                "PointingAlgorithms.set_boresight_to_look_at_geocentric_location, the target location is located within 10 cm of the observer. It is too close. It has been discarded"
            )
        return ok

    # ------------------------------------------------------------------------------
    #           get_unitvectors_at_target_location
    # ------------------------------------------------------------------------------
    def set_boresight_pointing_from_unitvectors(self, G: UnitVectors) -> bool:
        """
        Sets the rotation matrix of the platform to the given UnitVector array. The array defines the required x, y and
        z unit vectors in the :ref:`ecef`

        Parameters
        ----------
        G: UnitVectors
            The class that represents a (3x3) matrix. Each column of the matrix is the desired x,y,z unit vector
            in the :ref:`ecef` reference frame.

        Returns
        -------
        bool: True if successful.
        """
        self.platform.platform_pointing.force_pcf_rotation_matrix(G)
        return True

    # ------------------------------------------------------------------------------
    #           set_observer_to_look_in_azi_elev
    # ------------------------------------------------------------------------------
    def set_observer_to_look_in_azi_elev(
        self,
        observer: np.ndarray,
        elevation: float,
        azimuth: np.ndarray,
        roll_control: str,
        roll_angle_degrees: float,
    ) -> bool:
        """
        Sets the rotation matrix of the platform so the boresight of the instrument will look at the elevation
        and azimuth defined in the horizontal plane at the observers location.

        Parameters
        ----------
        observer: np.ndarray[3,]
            The x,y,z location of the observer in meters in the :ref:`ecef` reference frame.
        elevation: float
            The elevation in degrees of the look vector angle from the local horizontal plane at the observer.
            Horizontal is 0 degrees. vertical up is 90, vertical down is -90. It is a left handed rotation around the
            local west unit vector
        azimuth: float
            The azimuth in degrees of the look vector in the local horizontal plane of the observer. Azimuth is measured
            clockwise, like a compass, with 0 in the due North direction. 0=N, 90=E, 180=S, 270=W. It is a left handed rotation
            around the local up vector
        roll_control: str
            The string describing roll control, see: :ref:`rollcontrol`
        roll_angle_degrees: float
            The roll angle in degrees in the clockwise direction from the zero point set by `roll_control`

        Returns
        -------
        bool: True if successful.
        """
        self._geo.from_xyz(observer)
        up = self._geo.local_up
        north = -self._geo.local_south
        east = -self._geo.local_west
        bearing = math.radians(azimuth)
        theta = math.radians(elevation)
        horiz = math.cos(bearing) * north + math.sin(bearing) * east
        look = math.cos(theta) * horiz + math.sin(theta) * up
        G = self.apply_roll_control(
            look,
            observer,
            observer,
            roll_control,
            roll_angle_degrees,
            horizontalunitvector=horiz,
        )
        self.platform.platform_pointing.force_pcf_rotation_matrix(G)
        return True

    # ------------------------------------------------------------------------------
    #           config_roll_control
    # ------------------------------------------------------------------------------

    def apply_roll_control(
        self,
        lookvector: np.ndarray,
        target_location: np.ndarray,
        observer_location: np.ndarray,
        roll_cntl: str,
        roll_angle_degrees: float,
        horizontalunitvector: np.ndarray = None,
    ) -> UnitVectors:
        """
        Applys roll control to the look vector and returns the (3x3) array of Unit Vectors in the final :ref:`ecef` reference
        frame.

        Parameters
        ----------
        lookvector : np.ndarray(3,)
            Three element look unit vector away from the observer's position specified in the :ref:`ecef` coordinate system
        target_location : np.ndarray(3,)
            Three element position vector of the target location  specified in meters in the :ref:`ecef` coordinate system
        observer_location: np.ndarray(3,)
             Three element position vector of the observer's location specified in meters in the :ref:`ecef` coordinate system
        roll_cntl : str
            The class of required roll control. 'limb', 'nadir' and 'standard' are currently supported.
        roll_angle_degrees: float
            The required roll angle required using the current control frame.
        horizontalunitvector: np.ndarray(3,)
            Optional. This is the horizontal unit vector in the horizontal plane of the look vector. It is only used in standard
            roll control and allows the roll of the instrument unit vectors to stay well defined even when the looking straight up or down.

        Returns
        -------
        UnitVectors:
            Returns the 3x3 array of unit vectors required in the final :ref:`ecef` reference frame. The platform rotation
            matrix should be modified so that unit vectors in the :ref:`icf` will map to these unit vectors after rotation.
        """

        if roll_cntl == "limb":  # roll control is limb
            self._geo.from_xyz(
                target_location
            )  # so zero roll occurs when the Instrument Z axis, is as parallel as possible to the local up unit vector
            zunit = (
                self._geo.local_up
            )  # We want to choose a z unit vector. Get local up at the tangent point
            ythreshold = math.sin(math.radians(0.5))
            xunit = lookvector  # Get the look vector unit vector as the X direction
            yunit = np.cross(
                zunit, xunit
            )  # Get the y unit vector which points towards the left as seen from the oobserver.
            ymag = np.linalg.norm(
                yunit
            )  # Get the magnitude of the cross product. This becomes
            if (
                ymag < ythreshold
            ):  # if we are in an unstable region where the first choice of Z is too parallel to the look vector then use the second option  (ie  local Z axis and lookvector are with 0.57 degrees of each other)
                zunit2 = -self._geo.local_south  # so arbitrarily try local North
                yunit = np.cross(
                    zunit2, xunit
                )  # Get the y unit vector which points towards the left as seen from the oobserver.
                ymag = np.linalg.norm(
                    yunit
                )  # Get the magnitude of the cross product. This becomes
                yunit /= ymag  # get the magnitude of the Y unit vector
                zunit = np.cross(
                    xunit, yunit
                )  # now get the z unit vector perpendicular to the look vector and the y unit vector

        elif roll_cntl == "nadir":
            self._geo.from_xyz(
                target_location
            )  # so zero roll occurs when the Instrument Z axis, is as parallel as possible to the local north
            zunit = (
                -self._geo.local_south
            )  # We want to choose a z unit vector. Get local north at the target location
            ythreshold = math.sin(math.radians(0.5))
            xunit = lookvector  # Get the look vector unit vector as the X direction
            yunit = np.cross(
                zunit, xunit
            )  # Get the y unit vector which points towards the left as seen from the oobserver.
            ymag = np.linalg.norm(
                yunit
            )  # Get the magnitude of the cross product. This becomes
            if (
                ymag < ythreshold
            ):  # if we are in an unstable region where the first choice of Z is too parallel to the look vector then use the second option  (ie  local Z axis and lookvector are with 0.57 degrees of each other)
                zunit2 = -self._geo.local_up  # so arbitrarily use local north.
                yunit = np.cross(
                    zunit2, xunit
                )  # Get the y unit vector which points towards the left as seen from the oobserver.
                ymag = np.linalg.norm(
                    yunit
                )  # Get the magnitude of the cross product. This becomes
                yunit /= ymag  # get the magnitude of the Y unit vector
                zunit = np.cross(
                    xunit, yunit
                )  # now get the z unit vector perpendicular to the look vector and the y unit vector

        elif roll_cntl == "standard":
            self._geo.from_xyz(
                observer_location
            )  # so zero roll occurs when the Instrument Z axis, is as parallel as possible to the local up unit vector
            xunit = lookvector  # get the instrument x unit vector
            if (
                horizontalunitvector is not None
            ):  # If the caller has passed in the horizontal unit vector of the line of sight
                zunit = (
                    self._geo.local_up
                )  # then we can use it with good stability, so get the z unit vector
                yunit = np.cross(
                    zunit, horizontalunitvector
                )  # the zero point for the instrument y unit vector is in the horizontal plane perpendicular to the look vector. This always works even when the look vector is straight up.
                zunit = np.cross(
                    xunit, yunit
                )  # The zero point for the instrument z unit vector is the cross product of the x and y instrument unit vectors
            else:  # if the user cannot provide a horizontal bearing
                zunit = (
                    self._geo.local_up
                )  # then we try our best. We want to choose a z unit vector. Get local up at the tangent point
                z = np.dot(
                    lookvector, zunit
                )  # Get the vertical component of the look vector
                horiz = (
                    lookvector - z * zunit
                )  # And subtract it to get the horizontal component
                if (
                    math.fabs(z) > 0.9999619
                ):  # If we have almost no horizontal component( within 0.5 degrees of straight up/down)
                    horiz = (
                        -self._geo.local_south
                    )  # then arbitrarily use the north direction
                horiz = horiz / np.linalg.norm(
                    horiz
                )  # get the horizontal unit vector parallel to the horizontal component of the look vector
                yunit = np.cross(zunit, horiz)  # now get the instrument y unit vector
                zunit = np.cross(xunit, yunit)  # and the instrument z unit vector.
        else:
            msg = f"Unsupported value {roll_cntl} of roll control"
            raise ValueError(msg)
        if (
            roll_angle_degrees != 0.0
        ):  # Positive roll angle is clockwise as seen from theobserver looking along the line of sight
            theta = math.radians(
                -roll_angle_degrees
            )  # Apply the positive roll if the user has set it to non zero value
            costheta = math.cos(theta)
            sintheta = math.sin(theta)
            zprime = zunit * costheta + yunit * sintheta
            yprime = zunit * (-sintheta) + yunit * costheta
            zunit = zprime
            yunit = yprime
        return UnitVectors(vectors=(xunit, yunit, zunit))


# -----------------------------------------------------------------------------
#           OrientationTechniques
# -----------------------------------------------------------------------------
class OrientationTechniques:
    """
    A helper class that allows users to specify position and orientations using a set of position techniques
    and orientation techniques.
    """

    def __init__(self):
        """
        Creates a new empty instance.
        """
        self._position_convertors: dict[
            str,
            tuple[Callable[[Platform, np.datetime64, np.ndarray], bool], list[int]],
        ] = (
            {}
        )  # A dictionary of positioning tecniques. Each entry of the form [name, function]
        self._platform_orientation_convertors: dict[
            str, tuple[Callable[[PointingAlgorithms, np.ndarray, str], bool], list[int]]
        ] = (
            {}
        )  # A dictionary of orientation techniques. Each entry of the form [name, function]
        self._icf_look_vector_convertors: dict[
            str,
            tuple[Callable[[Platform, np.datetime64, np.ndarray], bool], list[int]],
        ] = (
            {}
        )  # A dictionary of ICF look vector techniques. Each entry of the form [name, function]
        self._position_definitions: list[
            tuple[str, np.ndarray | int]
        ] = (
            []
        )  # A list of various representations of position specifications. Each entry of the form [technique, data_array]
        self._platform_orientation_definitions: list[
            tuple[str, np.ndarray | int, str]
        ] = (
            []
        )  # A list of various representations of platform orientation techniques and data . Each entry of the form [technique, data_array, roll_control]
        self._icf_lookvector_definitions: list[
            tuple[str, np.ndarray | int, str]
        ] = []  # A list of various representations of ICF look vector specifications.
        self._utc: list[
            np.ndarray
        ] = []  # A list of arrays of measurement times of numpy.datetime64['us']
        self._isdirty = False

        self.add_position_convertor(
            "xyz", _technique_set_platform_position_from_xyz, [3]
        )
        self.add_position_convertor(
            "llh", _technique_set_platform_position_from_llh, [3]
        )
        self.add_position_convertor(
            "from_platform", _technique_set_platform_position_from_platform, [0]
        )
        self.add_position_convertor(
            "looking_at_llh",
            _technique_set_platform_position_from_observer_looking_at_llh,
            [5],
        )

        self.add_platform_orientation_convertor(
            "tangent_xyz_look", _technique_set_limb_look_vectors_from_unit_xyz, [3, 4]
        )  # Set look vectors using ITRF/ecef xyz location vectors
        self.add_platform_orientation_convertor(
            "tangent_altitude",
            _technique_set_look_vectors_from_tangent_altitude,
            [2, 3],
        )  # specifies
        self.add_platform_orientation_convertor(
            "tangent_from_orbitplane",
            _technique_set_boresight_look_at_location_orbitangle,
            [2, 3],
        )
        self.add_platform_orientation_convertor(
            "location_xyz", _technique_set_boresight_look_at_location_xyz, [3, 4]
        )
        self.add_platform_orientation_convertor(
            "location_llh", _technique_set_boresight_look_at_location_llh, [3, 4]
        )
        self.add_platform_orientation_convertor(
            "unit_vectors", _technique_set_boresight_pointing_from_unitvectors, [6]
        )
        self.add_platform_orientation_convertor(
            "azi_elev", _technique_set_observer_to_look_in_azi_elev, [2, 3]
        )
        self.add_platform_orientation_convertor(
            "yaw_pitch_roll", _technique_set_observer_to_look_in_azi_elev, [2, 3]
        )
        self.add_platform_orientation_convertor(
            "from_platform", _technique_set_platform_pointing_from_platform, [0]
        )

        self.add_icf_look_vector_convertor(
            "azi_elev", _technique_set_icf_orientation_from_azi_elev, [2]
        )
        self.add_icf_look_vector_convertor(
            "xyz", _technique_set_icf_look_from_xyz, [1, 3]
        )
        self.add_icf_look_vector_convertor(
            "nop", _technique_set_icf_orientation_no_operation, [0]
        )

    # ------------------------------------------------------------------------------
    #           add_platform_orientation_convertor
    # ------------------------------------------------------------------------------
    def add_platform_orientation_convertor(
        self,
        orientationtype: str,
        convertor_function: Callable[[PointingAlgorithms, np.ndarray, str], bool],
        num_user_params: list[int],
    ):
        key = orientationtype.lower()
        self._platform_orientation_convertors[key] = (
            convertor_function,
            num_user_params,
        )

    # ------------------------------------------------------------------------------
    #           add_position_convertor
    # ------------------------------------------------------------------------------
    def add_position_convertor(
        self,
        positiontype: str,
        convertor_function: Callable[[Platform, np.datetime64, np.ndarray], bool],
        num_user_params: list[int],
    ):
        key = positiontype.lower()
        self._position_convertors[key] = (convertor_function, num_user_params)

    # ------------------------------------------------------------------------------
    #           add_position_convertor
    # ------------------------------------------------------------------------------
    def add_icf_look_vector_convertor(
        self,
        lookvectorype: str,
        convertor_function: Callable[[Platform, np.datetime64, np.ndarray], bool],
        num_user_params: list[int],
    ):
        key = lookvectorype.lower()
        self._icf_look_vector_convertors[key] = (convertor_function, num_user_params)

    # -----------------------------------------------------------------------------
    #           numsamples
    # -----------------------------------------------------------------------------
    def num_measurements(self) -> int:
        """
        Returns the number of samples in this observation policy

        Returns
        -------
        int
            The number of samples in the observation policy.
        """
        n = 0
        for entry in self._utc:
            n += entry.size
        return n

    # -----------------------------------------------------------------------------
    #           clear
    # -----------------------------------------------------------------------------
    def clear(self):
        """
        Clears the current list of platform states inside the observation policy. This method should be called
        before starting a new observation policy
        """
        self._position_definitions.clear()
        self._platform_orientation_definitions.clear()
        self._icf_lookvector_definitions.clear()
        self._utc.clear()

    # ------------------------------------------------------------------------------
    #           check_position_converter_key
    # ------------------------------------------------------------------------------
    def _check_position_converter_key(self, key: str) -> bool:
        ok = self._position_convertors.get(key) is not None
        if not ok:
            msg = "The position converter keyword <{:s}> is not in the position converter dictionary. You will need to manually add this function with method OrientationTechniques.add_position_convertor".format(
                key
            )
            raise ValueError(msg)
        return ok

    # ------------------------------------------------------------------------------
    #           _check_platform_orientation_converter_key
    # ------------------------------------------------------------------------------
    def _check_platform_orientation_converter_key(self, key: str) -> bool:
        ok = self._platform_orientation_convertors.get(key) is not None
        if not ok:
            msg = "The platform orientation converter keyword <{:s}> is not in the platform orientation converter dictionary. You will need to manually add this function with method OrientationTechniques.add_platform_orientation_convertor".format(
                key
            )
            raise ValueError(msg)
        return ok

    # ------------------------------------------------------------------------------
    #           _check_icf_look_vector_converter_key
    # ------------------------------------------------------------------------------
    def _check_icf_look_vector_converter_key(self, key: str) -> bool:
        ok = self._icf_look_vector_convertors.get(key) is not None
        if not ok:
            msg = "The ICF look vector converter keyword <{}> is not in the platform orientation converter dictionary. You will need to manually add this function with method OrientationTechniques.add_icf_look_vector_convertor".format(
                key
            )
            raise ValueError(msg)
        return ok

    # ------------------------------------------------------------------------------
    #           _coerce_to_vector
    # ------------------------------------------------------------------------------
    def _coerce_to_vector(
        self, data: np.ndarray | Sequence[float], acceptable_n: list[int]
    ) -> np.ndarray:
        if isinstance(data, Sequence):
            data = np.array(data, dtype=np.float64)
        if type(data) is np.ndarray:
            if data.dtype != np.float64:
                data = np.array(data, dtype=np.float64)
            s = data.shape
            N = s[len(s) - 1]
            ok = N in acceptable_n
            if not ok:
                msg = "Incoming data must be an array or sequence with a trailing dimension being one of {}. Your array shape was {}".format(
                    acceptable_n, s
                )
                raise ValueError(msg)
            answer = np.copy(np.reshape(data, [data.size // N, N]))
        else:
            answer = None
            if data is not None:
                msg = "Incoming data must be a numpy array, sequence or None"
                raise ValueError(msg)
        return answer

    # ------------------------------------------------------------------------------
    #           _decode_position_entry
    # ------------------------------------------------------------------------------
    def _decode_position_entry(self, observer_positions):
        if isinstance(observer_positions, str):  # if we have just a string
            key = observer_positions  # then it is the technique key. It should be 'from_platform'
            data = None  # and there is no position data.
        else:  # otherwise it must be a sequence
            key = observer_positions[0]  # where the technique key is the first element
            data = (
                observer_positions[1] if len(observer_positions) > 1 else None
            )  # and the data sequence is the second element, or None if there is no second element
        key = key.lower()  # Convert the technique key to lower case
        if key == "from_platform":
            data = None
        self._check_position_converter_key(
            key
        )  # check that the technique key is supported
        num_params = self._position_convertors[key][1]
        positiondata = self._coerce_to_vector(
            data, num_params
        )  # coerce the vector data into array[N,3]
        numpositions = positiondata.shape[0] if (positiondata is not None) else 0
        return numpositions, positiondata, key

    # ------------------------------------------------------------------------------
    #           _decode_platform_orientation_entry
    # ------------------------------------------------------------------------------
    def _decode_platform_orientation_entry(self, platform_orientation_data):
        roll_control = "undefined"  # default value of roll if no value passed in
        data = None  # default value of data is no data are passed in
        if isinstance(platform_orientation_data, str):
            key = platform_orientation_data
        else:  # otherwise a dictionary has been passed
            key = platform_orientation_data[
                0
            ]  # get the technique key from the first element
            if len(platform_orientation_data) > 1:
                roll_control = platform_orientation_data[1]
                data = platform_orientation_data[2]
            else:
                data = None  # get the technique parameter data from the second element. None if there is no second element.
                roll_control = "undefined"
        key = key.lower()  # convert the technique key to lower case
        if key == "from_platform":
            data = None
            roll_control = "undefined"
        self._check_platform_orientation_converter_key(
            key
        )  # check that the look-vector technique is supported
        num_params = self._platform_orientation_convertors[key][
            1
        ]  # get the number of parameters expected from this look vector technique
        lookdata = self._coerce_to_vector(
            data, num_params
        )  # coerce the technique parameter data into a 2d-array of [num_params,N]
        numlooks = lookdata.shape[0] if (lookdata is not None) else 0
        return numlooks, lookdata, key, roll_control

    # ------------------------------------------------------------------------------
    #           _decode_instrument_internal_rotation
    # ------------------------------------------------------------------------------
    def _decode_icf_look_vector_entry(self, icf_look_vectors):
        data = None  # default value of data is no data are passed in
        if isinstance(icf_look_vectors, str):
            key = icf_look_vectors
        else:  # otherwise a dictionary has been passed
            if icf_look_vectors is not None:
                key = icf_look_vectors[
                    0
                ]  # get the technique key from the first element
                if len(icf_look_vectors) > 1:
                    data = icf_look_vectors[1]
            else:
                data = None  # get the technique parameter data from the second element. None if there is no second element.
                key = "nop"

            self._check_icf_look_vector_converter_key(
                key
            )  # check that the look-vector technique is supported
            num_params = self._icf_look_vector_convertors[key][
                1
            ]  # get the number of parameters expected from this look vector technique
            lookdata = self._coerce_to_vector(
                data, num_params
            )  # coerce the technique parameter data into a 2d-array of [num_params,N]
            numlooks = lookdata.shape[0] if (lookdata is not None) else 1
        return numlooks, lookdata, key

    # ------------------------------------------------------------------------------
    #           _decode_instrument_internal_rotation
    # ------------------------------------------------------------------------------
    def _decode_instrument_internal_rotation(self, instrument_internal_rotation):
        turntabledata = self._coerce_to_vector(
            instrument_internal_rotation, [2, 3]
        )  # coerce the technique parameter data into a 2d-array of [num_params,N]
        if turntabledata is None:  # If no turntable data are passed in
            turntabledata = np.zeros(
                [1, 3]
            )  # then create a 1 measurement array full of zeros
        numturntable = turntabledata.shape[
            0
        ]  # get the number of measurements from the first dimension
        numv = turntabledata.shape[1]  # get the number of parameters
        if numv != 3:  # if its not 3
            values = np.zeros(
                [numturntable, 3]
            )  # then resize the array to 3 with the last element equal to zero
            values[:, 0] = turntabledata[:, 0]  # copy over the first set of parameters
            values[:, 1] = turntabledata[:, 1]  # and the second set
            turntabledata = values  # and assign to the turntabledata
        return numturntable, turntabledata  # and we are done.

    # ------------------------------------------------------------------------------
    #           add_measurement_set
    # ------------------------------------------------------------------------------
    def add_measurement_set(
        self,
        utc: float
        | (
            str
            | (
                np.datetime64
                | (
                    datetime
                    | (
                        np.ndarray
                        | Sequence[
                            float | (str | (np.datetime64 | (datetime | np.ndarray)))
                        ]
                    )
                )
            )
        ),
        platform_position: tuple[str, Sequence | np.ndarray],
        platform_orientation: tuple[str, str, Sequence | np.ndarray],
        icf_orientation: tuple[str, str, Sequence | np.ndarray] | None = None,
    ):
        """
        Adds a set of measurements to the parent :class:~.Instrument`. The call adds **N** distinct measurements with
        **N* distinct settings of *time*, *position* and *orientation*. For convenience, settings for one variable can be represented
        with one value and it will be automatically expanded to the number of measurements represented in the other variables, i.e.
        borrowing the idea of **broadcasting** from *numpy*. Multiple measurement sets can be added to the :class:`~.Instrument`

        Parameters
        ----------
        utc: Array, sequence or scalar of string, float, datetime or datetime64
            The universal times of the set of measurements. It can be an array, sequence or scalar and each element
            should be a representation of universal time. If it is a scalar value then the same time value is assigned to every
            position given in `observer_positions`. All elements of the array should be of the same type and should represent
            universal time.
        roll_control: str
            The technique used to specify the location of zero roll in the :ref:`ecef` system. A description of values can
            be found in :ref:`rollcontrol`.
        platform_position: Tuple[str, sequence]
            A two element tuple. The first element of the tuple is a string that specifies the positioning technique. The second element
            of the tuple is a sequence or array (or None) that contains parameter data for the selected technique. The array should be of size (Npts, NParam)
            where  *Npts* is the number of points and *Nparam* is the number of parameters for the technique, e.g. (N,3), in the measurement set. The
            second element may be None if the technique is `from_platform`.
        platform_orientation:
            A three element tuple. The first element of the tuple is a string that specifies the orientation technique. The second element
            of the tuple is a sequence or array (or None) that contains parameter data for the selected technique. The array should be of size (Npts, NParam)
            where  *Npts* is the number of points and *Nparam* is the number of parameters for the technique, e.g. (N,3), in the measurement set. The
            second element may be None if the technique is `from_platform`.
        icf_orientation:
            Optional. A two element tuuple.  This option allows used to specify an array of instantaneous look vectors in the :ref:`icf`.
            The first element of the tuple is a string that specifies the ref:`icf` look vector  technique. The second element
            of the tuple is a sequence or array (or None) that contains parameter data for the selected technique. The array should be of size (Numlookvector, NParam)
            where  *Numlookvector* is the number of instantaneous look vectors points and *Nparam* is the number of parameters for the look vector technique, e.g. (N,3),
        """

        utdata = ut_to_datetime64(
            utc
        )  # Convert the incoming array of times to datetime64 for numpy
        if isinstance(
            utdata, np.datetime64
        ):  # If a scalar value comes back then we resize the times to math the positions
            utdata = np.array([utdata])  # make it into an array

        numtimes = utdata.size  # Get the number of time
        numpositions, positiondata, positionkey = self._decode_position_entry(
            platform_position
        )  # Get the number of positions, the key
        (
            numorientations,
            orientationdata,
            orientkey,
            roll_control,
        ) = self._decode_platform_orientation_entry(platform_orientation)
        numicflook, icflookdata, icfkey = self._decode_icf_look_vector_entry(
            icf_orientation
        )
        maxpts = np.max((numtimes, numpositions, numorientations, numicflook))

        if numtimes == 1:
            utdata = np.resize(
                utdata, [maxpts]
            )  # then resize the times array by simply resizing the current array
            numtimes = maxpts  # and reset the number of times
        else:
            if numtimes != maxpts:
                msg = "The number of time parameters given {} does not equal 1 or the number of looks {} and  and orientations given {}".format(
                    numtimes, numpositions, numorientations
                )
                raise ValueError(msg)

        if (
            positionkey == "from_platform"
        ):  # if the technique parameter array returns as None
            positiondata = maxpts  # set the technique parameter data as the number of times, used when passed to the technique algorithm
            numpositions = (
                maxpts  # and add the number of times to the number of positions
            )
        elif numpositions == 1:  # otherwise we have valied technique parameter data
            positiondata = np.tile(positiondata, [maxpts, 1])
            numpositions = (
                maxpts  # so get the number of points as the first element of the shape
            )
        else:
            if numpositions != maxpts:
                msg = "The number of position parameters given {} does not equal 1 or the number of times {} and orientations given {}".format(
                    numpositions, numtimes, numorientations
                )
                raise ValueError(msg)

        if (
            orientkey == "from_platform"
        ):  # if the technique parameter array returns as None
            orientationdata = maxpts  # set the technique parameter data as the number of times, used when passed to the technique algorithm
        elif numorientations == 1:  # otherwise we have valied technique parameter data
            orientationdata = np.tile(orientationdata, [maxpts, 1])
        else:
            if numorientations != maxpts:
                msg = "The number of orientation parameters given {} does not equal 1 or the number of times {} and positions {} given".format(
                    numorientations, numtimes, numpositions
                )
                raise ValueError(msg)

        if (numicflook == 1) and (icflookdata is not None):
            icflookdata = np.tile(icflookdata, [maxpts, 1])
        self._utc.append(
            utdata
        )  # Add the UTC times now that they are finalized in size
        self._position_definitions.append(
            (positionkey, positiondata)
        )  # if it is then append the definitions to the internal measurement set within this object
        self._platform_orientation_definitions.append(
            (orientkey, orientationdata, roll_control)
        )  # append these look vectors definitions to the internal measurement set within thgis object.
        self._icf_lookvector_definitions.append((icfkey, icflookdata))
        self._isdirty = True

    # ------------------------------------------------------------------------------
    #           make_observation_set
    # ------------------------------------------------------------------------------
    def make_observation_set(self, platform: Platform):
        """
        Converts the internal list of measurements from previous calls to :meth:`~.add_measurement_set` into an
        an internally cached :class:`~.PositionAndOrientationArray`. This method is normally called by the :class:`~.Platform` class.
        The internal list of measurement sets has only been cached up to this point. This code goes through the measurements
        and generates all the necessary platform positions and rotation matrices using all the specified techniques.

        Parameters
        ----------
        platform : Platform
            The :class:`~.Platform` object that will be updated with the measurements defined in this object.
        """
        pointingalgorithms = PointingAlgorithms(platform)
        postype = "----"
        orientationtype = "----"
        ilookseg = -1
        numlook = -1
        ilook = 0
        iposseg = -1
        numpos = -1
        ipos = 0
        iutc = 0
        iturn = 0
        numturn = -1
        iturnseg = -1
        allok = True

        for iseg in range(
            len(self._utc)
        ):  # Loop through all of the time segements. There is usually just one
            currentutc = self._utc[iseg]  # Get the current array of universal times
            numutc = (
                currentutc.size
            )  # Get the number of universal times in this segment
            for it in range(
                numutc
            ):  # Loop over all of the time values, we have already validated that the total number of positions and look arrays are the same size as the time array
                iutc += 1  # increase the time index to the next time measurement in the set of segments
                utc = currentutc[it]  # get the time. This should be np.datetime64

                iturn += 1  # Step to the next position entry in the current segment of internal turntable orientation entries
                if (
                    iturn >= numturn
                ):  # if we have exceeded the bounds of the current segment
                    iturnseg += 1  # then step to the next segment (it whould work as we have already validated data sets)
                    icflooktype, currentturn = self._icf_lookvector_definitions[
                        ilookseg
                    ]  # get the techniquename  to calculate ICF look vectors and the parameter data for this setting
                    icflookfunc = self._icf_look_vector_convertors[icflooktype][
                        0
                    ]  # get the ICF look vector technique function
                    iturn = 0  # Reset our position in the current segment
                    assert (currentturn is None) or (
                        type(currentturn) is np.ndarray
                    )  # We must have an array. There is no "from_platform" option
                    numturn = (
                        currentturn.shape[0] if currentturn is not None else 1
                    )  # then get the number of positions in this segment
                thisturn = (
                    currentturn[iturn, :] if currentturn is not None else None
                )  # get the technique parameter data for this measurement point.
                ok3 = icflookfunc(
                    platform, thisturn
                )  # Call the positioning technique function. Its sets the Platform position State. It returns false if it fails.

                ipos += 1  # Step to the next position entry in the current segment of position entries
                if (
                    ipos >= numpos
                ):  # if we have exceeded the bounds of the current segment
                    iposseg += 1  # then step to the next segment (it whould work as we have already validated data sets)
                    postype, currentpos = self._position_definitions[
                        iposseg
                    ]  # get the name of the position technique and the current position technique data for this segment
                    posfunc = self._position_convertors[postype][
                        0
                    ]  # Get the function we should call fo rthis technique
                    ipos = 0  # Reset our position in the current segment
                    if (
                        type(currentpos) is np.ndarray
                    ):  # If the position technique parameter is an array
                        numpos = currentpos.shape[
                            0
                        ]  # then get the number of positions ion this segment
                    else:  # otherwise its the special case, e.g. 'from_platform'
                        numpos = currentpos  # and the number of points is internally stored as the technique parameter data
                        currentpos = None  # and the real parameter data is None

                thispos = (
                    currentpos[ipos, :] if currentpos is not None else None
                )  # get the technique parameter data for this measurement point.
                ok1 = posfunc(
                    platform, utc, thispos
                )  # Call the positioning technique function. Its sets the Platform position State. It returns false if it fails.

                ilook += 1  # Step to the next position entry in the current list of orientation technique entries
                if (
                    ilook >= numlook
                ):  # if we have exceeded the bounds of the current segment
                    ilookseg += (
                        1  # then step to the next segment. It should always work
                    )
                    (
                        orientationtype,
                        currentlook,
                        roll_control,
                    ) = self._platform_orientation_definitions[
                        ilookseg
                    ]  # get the technique, parameter data and roll control for this settings
                    lookfunc = self._platform_orientation_convertors[orientationtype][
                        0
                    ]  # get the look technique function

                    ilook = 0
                    if type(currentlook) is np.ndarray:
                        numlook = currentlook.shape[0]
                    else:
                        numlook = currentlook
                        currentlook = None
                thislook = currentlook[ilook, :] if currentlook is not None else None
                ok2 = lookfunc(pointingalgorithms, thislook, roll_control)

                if ok1 and ok2 and ok3:
                    platform.add_current_platform_state()
                else:
                    logging.warning(
                        "Rejecting point {} from the observation policy. Positioning technique={} status = {}, Pointing Technique={} status = {}, Internal Rotation status = {}".format(  # noqa: G001
                            iutc, postype, ok1, orientationtype, ok2, ok3
                        )
                    )
                    allok = False
        if not allok:
            msg = "Platform.make_observation_set failed to create a valid observation set. Warning messages have been written to logging.warning"
            raise RuntimeError(msg)
        self._isdirty = False



from __future__ import annotations

import math

import numpy as np


# -----------------------------------------------------------------------------
#           class RotationMatrix:
# -----------------------------------------------------------------------------
class RotationMatrix:
    """
    A class that provides methods to calculate the rotation and transformation matrices used elsewhere in the package.
    These matrices are used to transform lines of sight and other vectors from one control frame to another.

        -  :py:meth:`from_transform_to_destination_coordinates`
        -  :py:meth:`from_rotate_one_setofaxes_to_another`
        -  :py:meth:`from_yaw_pitch_roll`
        -  :py:meth:`transform_vectors`

    """

    # ------------------------------------------------------------------------------
    #           IUnit
    # ------------------------------------------------------------------------------
    @staticmethod
    def IUnit():
        """
        Returns the identity matrix, :math:`\\boldsymbol{IUnit}` as a 3x3 matrix
        """
        return np.array(
            [
                [
                    1.0,
                    0.0,
                    0.0,
                ],  #: The rotation matiopn. By default it is the Unit matrix *"i.e. no effect"
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
            ],
            dtype="float64",
        )

    # -----------------------------------------------------------------------------
    #           __init__
    # -----------------------------------------------------------------------------

    def __init__(self, array: list[list[float]] | np.ndarray = None):
        self._R: np.ndarray = (
            RotationMatrix.IUnit()
            if array is None
            else np.array(array, dtype="float64")
        )  #: The rotation matiopn. By default it is the Unit Identity matrix *"i.e. no effect"

    # -----------------------------------------------------------------------------
    #           property R
    # -----------------------------------------------------------------------------
    @property
    def R(self):
        return self._R

    @property
    def RInv(self):
        return self._R.transpose()

    # -----------------------------------------------------------------------------
    #           from_rotation_matrix
    # -----------------------------------------------------------------------------
    def from_rotation_matrix(self, other: np.ndarray):
        """
        Sets the rotation from another 3x3 matrix
        """
        self._R = other

    # -----------------------------------------------------------------------------
    #           from_transform_to_destination
    # -----------------------------------------------------------------------------

    def from_transform_to_destination_coordinates(
        self, ax: np.ndarray, ay: np.ndarray, az: np.ndarray
    ):
        """
        Sets the rotation/transformation vector :math:`\\boldsymbol{R}` so it converts a vector, :math:`\\vec{v}`, specified in the system, :math:`\\boldsymbol{A}` (
        :math:`\\hat{x}_a` , :math:`\\hat{y}_a` , :math:`\\hat{z}_a` ),  to the same vector but specified in the destination system, :math:`\\boldsymbol{B}`
        (:math:`\\hat{x}_b` , :math:`\\hat{y}_b` , :math:`\\hat{z}_b`). This is implemented as a dot product.

        .. math::
               \\begin{split}
               \\vec{V}_A &= a_x\\hat{x}_a + a_y\\hat{y}_a + a_z\\hat{z}_a  \\\\
               \\vec{V}_B &= b_x\\hat{x}_b + b_y\\hat{y}_b + b_z\\hat{z}_b  \\\\
               \\end{split}

        where

        .. math::
               \\begin{split}
               \\hat{x}_a &=& a_{11}\\hat{x}_b +& a_{12}\\hat{y}_b +& a_{13}\\hat{z}_{b} \\\\
               \\hat{y}_a &=& a_{21}\\hat{x}_b +& a_{22}\\hat{y}_b +& a_{23}\\hat{z}_{b} \\\\
               \\hat{z}_a &=& a_{31}\\hat{x}_b +& a_{32}\\hat{y}_b +& a_{33}\\hat{z}_{b} \\\\
               \\end{split}

        then

        .. math::
                \\begin{pmatrix}
                Bx \\\\
                By \\\\
                Bz \\\\
                \\end{pmatrix}
                =
                \\begin{pmatrix}
                a_{11} & a_{21} & a_{31} \\\\
                a_{12} & a_{22} & a_{32} \\\\
                a_{13} & a_{23} & a_{33} \\\\
                \\end{pmatrix}
                \\begin{pmatrix}
                Ax  \\\\
                Ay  \\\\
                Az  \\\\
                \\end{pmatrix}

        Note that the 3x3 matrix is the *transpose* of the unit vector equations

        Parameters
        ----------
        ax
          A 3 element array expressing the :math:`\\hat{ax}` unit vector of frame :math:`\\boldsymbol{A}` in terms of the
          unit vectors in :math:`\\boldsymbol{B}`
        ay
          A 3 element array expressing the :math:`\\hat{ay}` unit vector of frame :math:`\\boldsymbol{A}` in terms of the
          unit vectors in :math:`\\boldsymbol{B}`
        az
          A 3 element array expressing the :math:`\\hat{az}` unit vector of frame :math:`\\boldsymbol{A}` in terms of the
          unit vectors in :math:`\\boldsymbol{B}`.

        """

        self._R[:, 0] = ax
        self._R[:, 1] = ay
        self._R[:, 2] = az

        # -----------------------------------------------------------------------------
        #           from_rotate_one_setofaxes_to_another
        # -----------------------------------------------------------------------------

    def from_rotate_one_setofaxes_to_another(self, A: np.ndarray, B: np.ndarray):
        """
        Generates the rotation matrix that rotates vectors specified in reference frame :math:`\\boldsymbol{A}` to vectors
        specified in reference frame :math:`\\boldsymbol{B}`. IUnit.E. the 3 unit vectors of frame :math:`\\boldsymbol{A}`  will be
        rotated on to the 3 unit vectors of frame  :math:`\\boldsymbol{B}`. The unit vectors for system :math:`\\boldsymbol{A}`, :math:`\\begin{bmatrix} \\hat{x}_A & \\hat{y}_A & \\hat{z}_A \\end{bmatrix}`
        and system :math:`\\boldsymbol{B}`, :math:`\\begin{bmatrix} \\hat{x}_B & \\hat{y}_B & \\hat{z}_B \\end{bmatrix}` are specified in a common
        system :math:`\\begin{bmatrix} \\hat{x'} & \\hat{y'} & \\hat{z'} \\end{bmatrix}` as 3x3 matrices.

        .. math::

                A= \\begin{pmatrix}
                   \\hat{x}_{A}.\\hat{x}' & \\hat{y}_{A}.\\hat{x}' & \\hat{z}_{A}.\\hat{x}' \\\\
                   \\hat{x}_{A}.\\hat{y}' & \\hat{y}_{A}.\\hat{y}' & \\hat{z}_{A}.\\hat{y}' \\\\
                   \\hat{x}_{A}.\\hat{z}' & \\hat{y}_{A}.\\hat{z}' & \\hat{z}_{A}.\\hat{z}' \\\\
                   \\end{pmatrix}

        .. math::

                B = \\begin{pmatrix}
                \\hat{x}_{B}.\\hat{x}' & \\hat{y}_{B}.\\hat{x}' & \\hat{z}_{B}.\\hat{x}' \\\\
                \\hat{x}_{B}.\\hat{y}' & \\hat{y}_{B}.\\hat{y}' & \\hat{z}_{B}.\\hat{y}' \\\\
                \\hat{x}_{B}.\\hat{z}' & \\hat{y}_{B}.\\hat{z}' & \\hat{z}_{B}.\\hat{z}' \\\\
                \\end{pmatrix}

        The rotation matrix :math:`\\boldsymbol{R}` is the operation that transforms the three unit vectors of :math:`\\boldsymbol{A}` in to
        the three unit vectors of :math:`\\boldsymbol{B}`

        .. math::

                \\boldsymbol{RA} &=  \\boldsymbol{B}\\\\
                \\boldsymbol{R}  &=  \\boldsymbol{BA^{-1}}\\\\

        Parameters
        ----------
        A
            A 3x3 matrix with the 3 unit vectors of control frame :math:`\\boldsymbol{A}` :math:`\\begin{bmatrix} \\hat{x}_A & \\hat{y}_A & \\hat{z}_A \\end{bmatrix}`
            expressed in a common coordinate sytem, :math:`\\hat{x}'`, :math:`\\hat{y}'`, :math:`\\hat{z}'`.

        B
            A 3x3 matrix with the 3 unit vectors of control frame :math:`\\boldsymbol{B}` :math:`\\begin{bmatrix} \\hat{x}_B & \\hat{y}_B & \\hat{z}_B \\end{bmatrix}`
            expressed in a common coordinate sytem, :math:`\\hat{x}'`, :math:`\\hat{y}'`, :math:`\\hat{z}'`.

        Examples
        --------
        Rotate Platform and Instrument to look at point on Earth
            A typical usage example is to calculate the rotation matrix for a platform so an instrument on that platform
            will look at a point on the Earth. The solution is to align the platform control frame with the local geodetic
            system (west, south, up) and then express the instrument control frame, :math:`\\boldsymbol{A}` , as geographic
            geocentric unit vectors via the west,south, up unit vectors.  The :math:`\\boldsymbol{B}` unit vectors are taken
            as the look unit vector from the platform to the desired point on Earth, :math:`\\hat{x}_B`, while
            :math:`\\hat{y}_B` and :math:`\\hat{z}_B` are chosen to meet observation requirements, eg horizontal slit etc.
            In this example both :math:`\\boldsymbol{A}` and :math:`\\boldsymbol{B}` can be readily expressed in geographic
            geocentric coordinates using other parts of the software package.

        Rotate Instrument Control Frame to Platform Control Frame
            For this example we want to rotate vectors specified in the instrument control frame so they are specified
            in the platform control frame. For example the CATS instrument was located on the backside of the gondola
            during the 2018 campaign and required the line of sight to be rotated so it looked backwards. The solution
            is specify the instrument control frame unit vectors in platform control frame coordinates and recognise that the
            common coordinates are the platform control frame unit vectors. In this case the instrument control frame is system
            :math:`\\boldsymbol{A}` and the platform control frame is system :math:`\\boldsymbol{B}`. Because system :math:`\\boldsymbol{B}`
            is the common system then matrix :math:`\\boldsymbol{B }` is simply the unit vector :math:`\\boldsymbol{IUnit}`.
            Matrix :math:`\\boldsymbol{A}` must be calculated from instrument and platform configuration details but
            then :math:`\\boldsymbol{R}` is readily derived.


        """

        self._R = B @ np.linalg.inv(
            A
        )  # Note that you can also use the transpose of A, as it should be the inverse for proper rotations

    # -----------------------------------------------------------------------------
    #           from_yaw_pitch_roll
    # -----------------------------------------------------------------------------

    def from_yaw_pitch_roll(self, yaw: float, pitch: float, roll: float):
        """
        Creates a rotation matrix that applies yaw, pitch and roll in that order. This method is intended for aircraft
        system where :math:`\\hat{x}` is looking forward looking, :math:`\\hat{z}` is downwards and :math:`{y}` is towards starboard.
        All angles are made in a right-handed sense around the axis of rotation. Typically, positive pitch is turning "*upwards*",
        positive yaw is "*turning towards the right*" and positive roll is "*leaning to the right*".

        The matrix is created in the rotation order Yaw, Pitch, Roll (`rzyx`). The rotations are intrinsic, meaning the axes of
        rotations are also rotated. Intrinsic rotations require the rotations to be applied in reverse order (ie roll, pitch, yaw)
        but then the matrix operations are applied from right to left so we get the order yaw, @ pitch @ roll in the actual code to
        represent (yaw, pitch, roll).

        Parameters
        ----------
        yaw
            The yaw angle in radians. This rotation is applied first to the z axis.
        pitch
            The pitch angle in radians. This rotation is applied second to the rotated y axis.
        roll
            The roll angle in radians. This rotation is applied last to the rotated x axis.
        """

        yawMatrix = np.array(
            [
                [math.cos(yaw), -math.sin(yaw), 0],  # Yaw is rotation around the Z axis
                [math.sin(yaw), math.cos(yaw), 0],
                [0, 0, 1],
            ]
        )

        pitchMatrix = np.array(
            [
                [
                    math.cos(pitch),
                    0,
                    math.sin(pitch),
                ],  # Pitch is rotation about the Y axis
                [0, 1, 0],
                [-math.sin(pitch), 0, math.cos(pitch)],
            ]
        )

        rollMatrix = np.array(
            [
                [1, 0, 0],  # Roll is rotation around the X axis
                [0, math.cos(roll), -math.sin(roll)],
                [0, math.sin(roll), +math.cos(roll)],
            ]
        )

        self._R = (
            yawMatrix @ pitchMatrix @ rollMatrix
        )  # Apply rotation ZYX, (note it is an instrinsic rotation so we apply roll first then pitch then yaw)

    # -----------------------------------------------------------------------------
    #           from_azimuth_elevation_roll
    # -----------------------------------------------------------------------------

    def from_azimuth_elevation_roll(
        self, azimuthradians: float, elevationradians: float, rollradians: float
    ):
        """
        Create a rotation matrix that rotates a vector by the specified azimuth, elevation and roll. It is is suitable
        for systems where the :math:`\\hat{x}` is horizontal and forwards, the :math:`\\hat{z}` axis is upwards and the
        :math:`\\hat{y}` points to the portside (left).  In this system, azimuth is a left handed (clockwise, compass) rotation
        around :math:`\\hat{z}`, elevation is a left handed rotation around :math:`\\hat{y}` (up is positive) and roll is a right-handed
        rotation around :math:`\\hat{x}` (leaning right is positive). The rotation matrix is created in the rotation order Azimuth,
        Elevation, Roll (`rzyx`)

        Parameters
        ----------
        azimuthradians
            The azimuthual rotation in radians. The rotation is a left-handed rotation around the :math:`\\hat{z}` axis.

        elevationradians
            The vertical rotation in radians. The rotation is a left-handed rotation around the :math:`\\hat{y}` axis.

        rollradians
            The roll rotation in radians. The rotation is in a right haded sense around the :math:`\\hat{x}` axis

        """

        self.from_yaw_pitch_roll(-azimuthradians, -elevationradians, rollradians)

    # -----------------------------------------------------------------------------
    #           transform_vectors
    # -----------------------------------------------------------------------------

    def transform_vectors(self, lookvector: np.ndarray) -> np.ndarray:
        """
        Applies the internal rotation matrix to an array of N vectors.
        The incoming array must of the form (3,N) where the first dimension contains the X,Y,Z components of the N
        vectors.

        Parameters
        ----------
        lookvector:
            An array (3,N) of incoming vectors that will be rotated/transformed. The first dimension contains the X,Y,Z
            components of the N vectors

        Returns
        -------
        An array of size (3,N) which is the set of vectors generated by applying the rotation matrix to the incoming vectors.
        The first dimension contains the X,Y,Z components of the rotated N vectors

        """

        return self._R @ lookvector


# -----------------------------------------------------------------------------
#           class UnitVectors:
# -----------------------------------------------------------------------------
class UnitVectors:
    """
    A class to hold the unit vectors :math:`\\begin{bmatrix} \\hat{x} & \\hat{y} & \\hat{z} \\end{bmatrix}` for one
    reference frame specified in terms of another reference frame, :math:`\\begin{bmatrix} \\hat{x}' & \\hat{y}' & \\hat{z}' \\end{bmatrix}`
    The components of each unit vector occupy a column of the matrix.

        .. math::

             \\begin{pmatrix}
             \\hat{x}.\\hat{x}' & \\hat{y}.\\hat{x}' & \\hat{z}.\\hat{x}' \\\\
             \\hat{x}.\\hat{y}' & \\hat{y}.\\hat{y}' & \\hat{z}.\\hat{y}' \\\\
             \\hat{x}.\\hat{z}' & \\hat{y}.\\hat{z}' & \\hat{z}.\\hat{z}' \\\\
             \\end{pmatrix}

    Note that the components of :math:`\\begin{bmatrix} \\hat{x}' & \\hat{y}' & \\hat{z}' \\end{bmatrix}` in terms of
    :math:`\\begin{bmatrix} \\hat{x} & \\hat{y} & \\hat{z} \\end{bmatrix}` is the transpose (which is also the inverse)
    of this matrix.

    """

    def __init__(
        self,
        vectors: list[list[float]]
        | (tuple[np.ndarray, np.ndarray, np.ndarray] | np.ndarray) = None,
    ):
        self._R: np.ndarray = (
            RotationMatrix.IUnit() if vectors is None else np.array(vectors).transpose()
        )

    @property
    def R(self):
        return self._R

    def X(self) -> np.ndarray:
        return self._R[:, 0]

    def Y(self) -> np.ndarray:
        return self._R[:, 1]

    def Z(self) -> np.ndarray:
        return self._R[:, 2]

    # -----------------------------------------------------------------------------
    #           roll_pitch_yaw
    # -----------------------------------------------------------------------------

    def roll_pitch_yaw(self) -> tuple[float, float, float]:
        """
        Returns the yaw, pitch and roll in radians that respresents
        this rotation matrix

        Returns
        -------
        Tuple[float,float,float]
            A three element tuple of (yaw,pitch,roll). Each angle is expressed in radians.
        """

        r = self.R
        r11 = r[0, 0]
        r21 = r[1, 0]
        r31 = r[2, 0]
        r32 = r[2, 1]
        r33 = r[2, 2]
        sinbeta = -r31  # get (3,1)
        cosbeta = math.sqrt(r32 * r32 + r33 * r33)  # r[3,2]^2 + r[3,3]^2
        pitch = math.atan2(sinbeta, cosbeta)
        yaw = math.atan2(r21, r11)
        roll = math.atan2(r32, r33)
        return yaw, pitch, roll



from __future__ import annotations

import math
from collections import namedtuple

import numpy as np
from numpy.linalg import norm

# ------------------------------------------------------------------------------
#           ClassicalOrbitalElements
# ------------------------------------------------------------------------------
ClassicalOrbitalElements = namedtuple(
    "ClassicalOrbitalElements", ("H", "e", "RA", "i", "W", "TA", "a")
)
ClassicalOrbitalElements.__doc__ = """
    The 6 elements of classical orbital elements

    Parameters
    ----------
    H : float
        the angular momemtum (m^2/s)
    e : float
        eccentricity
    RA : float
        Right ascension of the ascending node in radians
    i : float
        inclination of the orbit
    W : float
        argument of perigee in radians
    TA : float
        True anomaly in radians
    a : float
        semimajor axis
    """


# ------------------------------------------------------------------------------
#           coe_from_state_vector
# ------------------------------------------------------------------------------
def coe_from_state_vector(R: np.ndarray, V: np.ndarray) -> ClassicalOrbitalElements:
    """
    Calculates classical orbital elements (ie Kepler) given a position and velocity. This is taken from
    Section 4.1 and Appendix D.8 Algorithm 4.1: of Orbital Mechanics for Engineering Students, Howard Curtis, 2005,
    ISBN 0 7506 6169 0.

    Parameters
    ----------
    R : np.ndarray(3)
        The position vector of the satellite in meters. This is typically expressed in ECI coordinates for Earth based
        satellites
    V : np.ndarray()
    mu - gravitational parameter (km^3/s^2)
    R - position vector in the geocentric equatorial frame (m)
    V - velocity vector in the geocentric equatorial frame (m/s)
    r, v - the magnitudes of R and V

    a - semimajor axis (km)
    pi - 3.1415926...
    coe - vector of orbital elements [h e RA incl w TA a]
    """

    mu = 3.986004418e14
    eps = 1.0e-10
    r = norm(R)
    v = norm(V)
    vr = np.dot(R, V) / r  # radial velocity component (m/s)
    H = np.cross(R, V)  # the angular momentum vector (m^2/s)
    h = norm(H)  # the magnitude of H (m^2/s)
    incl = math.acos(H[2] / h)  # Equation 4.7: inclination of the orbit (rad)
    N = np.cross([0, 0, 1], H)  # Equation 4.8: the node line vector (km^2/s)
    n = norm(N)  # the magnitude of N

    if n != 0:  # Equation 4.9: right ascension of the ascending node (rad)
        RA = math.acos(N[0] / n)
        if N[1] < 0:
            RA = 2 * math.pi - RA
    else:
        RA = 0

    E = (
        1.0 / mu * ((v * v - mu / r) * R - r * vr * V)
    )  # Equation 4.10: eccentricity vector
    e = norm(E)  # Calculate eccentricity

    if n != 0:  # Equation 4.12 (incorporating the case e = 0)
        if (
            e > eps
        ):  # eps - a small number below which the eccentricity is considered to be zero
            w = math.acos(np.dot(N, E) / n / e)  # w is argument of perigee (rad)
            if E[2] < 0:
                w = 2 * math.pi - w
        else:
            w = 0
    else:
        w = 0

    if e > eps:  # Equation 4.13a (incorporating the case e = 0):
        TA = math.acos(np.dot(E, R) / e / r)  # TA is true anomaly (rad)
        if vr < 0:
            TA = 2 * math.pi - TA
    else:
        cp = np.cross(N, R)
        if cp[2] >= 0:
            TA = math.acos(np.dot(N, R) / n / r)
        else:
            TA = 2 * math.pi - math.acos(np.dot(N, R) / n / r)

    a = (
        h * h / mu / (1.0 - e * e)
    )  # Equation 2.61, semimajor axis (m),  (a < 0 for a hyperbola)
    return ClassicalOrbitalElements(H=h, e=e, RA=RA, i=incl, W=w, TA=TA, a=a)



from __future__ import annotations

import math
from datetime import datetime

import numpy as np

import skretrieval.time.eci

from .satellitebase import SatelliteBase


# -----------------------------------------------------------------------------
#           SatelliteSimpleGeostationary
# -----------------------------------------------------------------------------
class SatelliteSimpleGeostationary(SatelliteBase):
    """
    Implements a simple geostationary satellite that simply stays above a fixed location on the equator.
    """

    def __init__(self, longitude_degrees: float):
        """
        Create a geostationary orbit that remains fixed over the the given longitude.

        Parameters
        ----------
        longitude_degrees:: float
            The geographic longitude of the geostationary orbit. expressed in degrees, +ve east.
        """

        super().__init__()
        radius = 42164000.0  # radius of orbit in meters
        theta = math.radians(longitude_degrees)
        omega = math.pi * 2.0 / self.period()  # angular velocaity radiancs.sec
        x = radius * math.cos(theta)
        y = radius * math.sin(theta)
        vx = -radius * omega * math.sin(theta)
        vy = radius * omega * math.cos(theta)
        self._geoposition = np.array(
            [x, y, 0.0]
        )  # The geocentric, geographic location of the satellite
        self._geovelocity = np.array(
            [vx, vy, 0.0]
        )  # The geocentric, geographic velocity of the satellite

    # -----------------------------------------------------------------------------
    #               def update_eci_position(self, platform_utc:datetime):
    # -----------------------------------------------------------------------------
    def update_eci_position(self, utc: datetime):
        """
        Not usually used by users. Updates the ECI location of the satellite.

        Parameters
        ----------
        utc : datetime
            Updates the ECI position and ECI velocity of the satellite to this Coordinated Universal Time
        """
        ecipos = skretrieval.time.eci.itrf_to_eciteme(self._geoposition, utc)
        ecivel = skretrieval.time.eci.itrf_to_eciteme(self._geovelocity, utc)
        self._set_current_state(ecipos, ecivel, utc)

    # ------------------------------------------------------------------------------
    #           period
    # ------------------------------------------------------------------------------
    def period(self) -> float:
        """
        Returns the period of the geostationary orbit to be exactly one sidereal day.

        Returns
        -------
        float
            The period of the satellite in seconds
        """
        return 86164.091

    # ------------------------------------------------------------------------------
    #           eccentricity
    # ------------------------------------------------------------------------------
    def eccentricity(self) -> float:
        """
        Returns the eccentricity of the orbit. It will always be zero.
        """
        return 0.0



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import xarray as xr
from scipy import sparse

from skretrieval.geodetic import target_lat_lon_alt


class RadianceBase(ABC):
    """
    Base functionality that every radiance format must support
    """

    def __init__(self, ds: xr.Dataset):
        self._ds = ds

    @property
    def data(self):
        return self._ds

    @data.setter
    def data(self, value):
        self._ds = value

    def tangent_locations(self):
        """
        Calculates tangent locations for all lines of sight.  If the line of sight does not have a tangent location
        the ground intersection is returned instead
        Returns
        -------
        xr.Dataset
            Dataset containing 'latitude', 'longitude', and 'altitude' of the tangent locations

        """
        los_dims = [dim for dim in self._ds["los_vectors"].dims if dim != "xyz"]

        stacked_los = self._ds["los_vectors"].stack(temp_dim=los_dims)  # noqa: PD013
        stacked_obs = self._ds["observer_position"].stack(  # noqa: PD013
            temp_dim=los_dims
        )

        latitudes = []
        longitudes = []
        altitudes = []
        for idx in stacked_los["temp_dim"]:
            one_los = stacked_los.sel(temp_dim=idx)
            one_obs = stacked_obs.sel(temp_dim=idx)

            lat, lon, alt = target_lat_lon_alt(one_los.to_numpy(), one_obs.to_numpy())

            latitudes.append(lat)
            longitudes.append(lon)
            altitudes.append(alt)

        result = xr.Dataset(
            {
                "latitude": (["temp_dim"], latitudes),
                "longitude": (["temp_dim"], longitudes),
                "altitude": (["temp_dim"], altitudes),
            },
            coords=stacked_los.coords,
        )

        return result.unstack("temp_dim").drop("xyz")  # noqa: PD010

    @abstractmethod
    def to_raw(self):
        pass


class RadianceRaw(RadianceBase):
    def __init__(self, ds: xr.Dataset):
        """
        Raw radiance measurements.  This is the simplest structure that can hold measurements from a wide variety of
        instruments.  Each measurement is represented by a line of sight (look vector, observer position, time) and
        a single radiance value.  Optionally, additional parameters may be added such as the estimated noise.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset containing the variables ['radiance', 'los_vector', 'observer_position'] and the coordinates ['meas']
            'meas' is an index that goes across all measurements (could be wavelength or los changes)
        """
        super().__init__(ds)

    def to_raw(self):
        return self

    def _validate_format(self):
        return True


class RadianceGridded(RadianceBase):
    def __init__(self, ds: xr.Dataset):
        """
        A specific radiance format where the radiance data can be represented on a (wavelength, line of sight) grid.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset containing the variable ['radiance', 'los_vector', 'observer_position'] and the coordinates
            ['wavelength', 'los'] where 'los' is a simple indexer that goes across changes in the line of sight.
        """
        super().__init__(ds)

    def _validate_format(self):
        return True

    def to_raw(self):
        new_ds = self.data.stack(meas=["wavelength", "los"])  # noqa: PD013

        return RadianceRaw(new_ds)


class RadianceSpectralImage(RadianceGridded):
    def __init__(self, ds, num_columns: int | None = None, num_rows: int | None = None):
        """
        A Specific radiance format to hold a (wavelength x columns x rows) grid of data. Only one of `num_columns` or
        `num_rows` should be specified.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset containing the variable ['radiance', 'los_vector', 'observer_position'] and the coordinates
            ['wavelength', 'los'] where 'los' is a simple indexer that goes across changes in the line of sight.
        num_columns : int
            Number of columns in the radiance grid. Optional, inferred from num_rows if not provided.
        num_rows : int
            Number of rows in the radiance grid. Optional, inferred from num_columns if not provided.
        """
        if (num_columns is None) and (num_rows is None):
            msg = "either num_columns or num_rows must be specified"
            raise ValueError(msg)

        if num_rows is None:
            num_rows = int(len(ds.los) / num_columns)
        elif num_columns is None:
            num_columns = int(len(ds.los) / num_rows)

        if num_rows * num_columns != len(ds.los):
            msg = "number of pixels must equal the number of lines of sight"
            raise ValueError(msg)

        nx = np.arange(0, num_columns)
        ny = np.arange(0, num_rows)
        mi = pd.MultiIndex.from_product([ny, nx], names=["ny", "nx"])
        dsc = ds.copy()
        dsc.coords["los"] = mi
        dsc = dsc.unstack("los")  # noqa: PD010

        super().__init__(dsc)


class RadianceOrbit:
    """
    A collection of other RadianceFormats that combine together to create an entire orbit of L1 data

    For example an entire orbit of OMPS data can be created either as a single RadianceGridded for the entire orbit
    or as a List of single RadianceGridded for each image
    """

    def __init__(
        self,
        data: list[RadianceBase],
        wf: list[sparse.spmatrix] | None = None,
        wf_names=None,
    ):
        self._data = data
        self._wf = wf
        self._wf_names = wf_names

        # Create the image slices
        self._slices = []
        cur_idx = 0
        for rad in self._data:
            self._slices.append(slice(cur_idx, cur_idx + len(rad.data.los)))
            cur_idx += len(rad.data.los)

    def derived_type(self):
        return type(self._data[0])

    @property
    def wf(self):
        return self._wf

    @property
    def wf_names(self):
        return self._wf_names

    def image_radiance(self, index, dense_wf=False):
        radiance = self._data[index]
        if self._wf is not None:
            if dense_wf:
                if self._wf_names is None:
                    wfs = np.array(
                        [wf[self._slices[index], :].toarray() for wf in self._wf]
                    )
                    radiance.data["wf"] = (["wavelength", "los", "perturbation"], wfs)
                else:
                    for wf_name, specieswf in zip(self._wf_names, self._wf):
                        wfs = np.array(
                            [wf[self._slices[index], :].toarray() for wf in specieswf]
                        )
                        radiance.data[wf_name] = (
                            ["wavelength", "los", "perturbation"],
                            wfs,
                        )
                return radiance
            if self._wf_names is None:
                wfs = np.array([wf[self._slices[index], :] for wf in self._wf])
                return radiance, wfs
            wfs = []
            for specieswf in self._wf:
                wfs.append(np.array([wf[self._slices[index], :] for wf in specieswf]))
            return radiance, wfs
        return radiance

    @property
    def num_images(self):
        return len(self._data)



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
from sasktran import Geometry

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry


class Sensor(ABC):
    """
    A sensor is an object which takes in radiances at the front of an aperture, and returns backend radiances
    observed by the sensor.
    """

    @abstractmethod
    def model_radiance(
        self,
        optical_geometry: OpticalGeometry,
        model_wavel_nm: np.array,
        model_geometry: Geometry,
        radiance: np.array,
        wf=None,
    ) -> radianceformat.RadianceBase:
        """
        Takes in high resolution radiances at the front of the aperture and converts them to what is observed by the
        sensor.

        Parameters
        ----------
        optical_geometry : OpticalGeometry
            The orientation, look vector, time, position of the sensor
        model_wavel_nm : np.array
            Array of radiances that correspond to radiance
        model_geometry : Geometry
            The Geometry object that corresponds to radiance
        radiance : np.array
            An array (wavelength, line of sight) of input radiances
        wf : np.array, optional
            An array (wavelength, line of sight, perturbation) corresponding to weighting functions. Default None.

        Returns
        -------
        radianceformat.RadianceBase
            Output L1 radiances in a format specific to the sensor.  The format is defined by `radiance_format`

        """

    @abstractmethod
    def measurement_geometry(self, optical_geometry: OpticalGeometry):
        """
        Takes in the sensors orientation and returns back a Geometry object corresonding to the central
        point of each measurement.

        Parameters
        ----------
        optical_geometry: OpticalGeometry
            Sensor orientation

        Returns
        -------
        Geometry
        """

    @abstractmethod
    def measurement_wavelengths(self):
        """

        Returns
        -------
        np.array
            Central wavelengths of the measurement
        """

    @staticmethod
    @abstractmethod
    def radiance_format():
        """
        Returns the format that model_radiance returns data in

        Returns
        -------
        A specialized instance of radianceformat.RadianceBase
        """



from __future__ import annotations

from copy import copy

import numpy as np
import xarray as xr

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import LineShape
from skretrieval.core.sasktranformat import SASKTRANRadiance
from skretrieval.core.sensor import Sensor


class Spectrograph(Sensor):
    def __init__(
        self,
        wavelength_nm: np.array,
        pixel_shape: list[LineShape],
        vert_fov: LineShape,
        horiz_fov: LineShape,
        spectral_native_coordinate: str = "wavelength_nm",
    ):
        """
        A spectrograph is a 1D array of pixels

        Parameters
        ----------
        wavelength_nm : np.array
            Central wavelengths for each pixel
        pixel_shape: LineShape
            Wavelength line shape
        vert_fov: LineShape
            Vertical field of view
        horiz_fov: LineShape
            Horizontal field of view
        """

        self._wavelength_nm = wavelength_nm
        self._wavenumber_cminv = 1e7 / wavelength_nm
        self._pixel_shape = pixel_shape

        self._vert_fov = vert_fov
        self._horiz_fov = horiz_fov

        self._cached_wavel_interp = None
        self._cached_wavel_interp_wavel = None

        self._spectral_native_coordinate = spectral_native_coordinate

    def _construct_interpolators(self, orientation, los_vectors, model_spectral_grid):
        x_axis = np.array(orientation.look_vector)
        vert_normal = np.cross(np.array(x_axis), np.array(orientation.local_up))
        vert_normal = vert_normal / np.linalg.norm(vert_normal)
        vert_y_axis = np.cross(vert_normal, x_axis)

        horiz_y_axis = vert_normal

        horiz_angle = []
        vert_angle = np.arctan2(
            np.dot(los_vectors, vert_y_axis), np.dot(los_vectors, x_axis)
        )

        horiz_angle = np.arctan2(
            np.dot(los_vectors, horiz_y_axis), np.dot(los_vectors, x_axis)
        )

        horiz_interpolator = self._horiz_fov.integration_weights(
            0, np.array(horiz_angle)
        )
        vert_interpolator = self._vert_fov.integration_weights(0, np.array(vert_angle))

        los_interpolator = np.zeros(len(vert_interpolator))
        los_interpolator = horiz_interpolator * vert_interpolator
        los_interpolator /= np.nansum(los_interpolator)

        los_interpolator = los_interpolator.reshape(-1, 1)

        if not np.array_equal(model_spectral_grid, self._cached_wavel_interp_wavel):
            wavel_interp = []
            for cw, p in zip(self._wavelength_nm, self._pixel_shape):
                weights = p.integration_weights(cw, model_spectral_grid)

                wavel_interp.append(weights / weights.sum())

            wavel_interp = np.vstack(wavel_interp)
            self._cached_wavel_interp = wavel_interp
            self._cached_wavel_interp_wavel = copy(model_spectral_grid)

        return self._cached_wavel_interp, los_interpolator

    def model_radiance(
        self,
        radiance: SASKTRANRadiance,
        orientation: OpticalGeometry,
    ) -> radianceformat.RadianceGridded:
        wavel_interp, los_interp = self._construct_interpolators(
            orientation,
            radiance.data["look_vectors"].to_numpy(),
            radiance.data["wavelength_nm"].to_numpy(),
        )

        modelled_radiance = np.einsum(
            "ij,jk...,kl",
            wavel_interp,
            radiance.data["radiance"].to_numpy(),
            los_interp,
            optimize=True,
        )

        data = xr.Dataset(
            {
                "radiance": (["wavelength", "los"], modelled_radiance),
                "mjd": (["los"], [orientation.mjd]),
                "los_vectors": (
                    ["los", "xyz"],
                    orientation.look_vector.reshape((1, 3)),
                ),
                "observer_position": (
                    ["los", "xyz"],
                    orientation.observer.reshape((1, 3)),
                ),
            },
            coords={
                "wavelength": self._wavelength_nm,
                "xyz": ["x", "y", "z"],
            },
        )
        for key in list(radiance.data):
            if key.startswith("wf"):
                modelled_wf = np.einsum(
                    "ij,ljk,km->iml",
                    wavel_interp,
                    radiance.data[key].to_numpy(),
                    los_interp,
                    optimize=True,
                )

                data[key] = (
                    ["wavelength", "los", radiance.data[key].dims[0]],
                    modelled_wf,
                )

        return radianceformat.RadianceGridded(data)

    def radiance_format(self) -> type[radianceformat.RadianceGridded]:
        return radianceformat.RadianceGridded


class SpectrographOnlySpectral(Sensor):
    def __init__(
        self,
        wavelength_nm: np.array,
        pixel_shape: list[LineShape],
        spectral_native_coordinate: str = "wavelength_nm",
        assign_coord: str = "wavelength",
    ):
        """
        Similar to a spectrograph but does not perform convolution in spatial space, just wavelength

        Parameters
        ----------
        wavelength_nm : np.array
            Central wavelengths for each pixel
        pixel_shape: LineShape
            Wavelength line shape
        """

        self._wavelength_nm = wavelength_nm
        self._wavenumber_cminv = 1e7 / wavelength_nm
        self._pixel_shape = pixel_shape

        self._cached_wavel_interp = None
        self._cached_wavel_interp_wavel = None

        self._spectral_native_coordinate = spectral_native_coordinate
        self._assign_coord = assign_coord

        if spectral_native_coordinate == "wavelength_nm":
            self._assign_vals = self._wavelength_nm
        else:
            self._assign_vals = self._wavenumber_cminv

    def _construct_interpolators(self, model_spectral_grid):
        if not np.array_equal(model_spectral_grid, self._cached_wavel_interp_wavel):
            wavel_interp = []
            for cw, p in zip(self._assign_vals, self._pixel_shape):
                weights = p.integration_weights(cw, model_spectral_grid)

                wavel_interp.append(weights / weights.sum())

            wavel_interp = np.vstack(wavel_interp)
            self._cached_wavel_interp = wavel_interp
            self._cached_wavel_interp_wavel = copy(model_spectral_grid)

        return self._cached_wavel_interp

    def model_radiance(
        self,
        radiance: SASKTRANRadiance,
        orientation: OpticalGeometry,  # noqa: ARG002
    ) -> radianceformat.RadianceGridded:
        wavel_interp = self._construct_interpolators(
            radiance.data[self._spectral_native_coordinate].to_numpy()
        )

        modelled_radiance = np.einsum(
            "ij,jk...",
            wavel_interp,
            radiance.data["radiance"].isel(stokes=0).to_numpy(),
            optimize=True,
        )

        data = xr.Dataset(
            {
                "radiance": ([self._assign_coord, "los"], modelled_radiance),
            },
            coords={
                self._assign_coord: self._assign_vals,
                "xyz": ["x", "y", "z"],
            },
        )

        if "look_vectors" in radiance.data:
            data["los_vectors"] = radiance.data["look_vectors"]

        if "observer_position" in radiance.data:
            data["observer_position"] = radiance.data["observer_position"]

        for key in list(radiance.data):
            if key.startswith("wf"):
                modelled_wf = np.einsum(
                    "ij,ljk->ikl",
                    wavel_interp,
                    radiance.data[key].isel(stokes=0).to_numpy(),
                    optimize=True,
                )

                data[key] = (
                    [self._assign_coord, "los", radiance.data[key].dims[0]],
                    modelled_wf,
                )

        return radianceformat.RadianceGridded(data)

    def radiance_format(self) -> type[radianceformat.RadianceGridded]:
        return radianceformat.RadianceGridded


def _set_join(lower_bounds, upper_bounds):
    final_sets = [[lower_bounds[0], upper_bounds[0]]]

    for lower, upper in zip(lower_bounds, upper_bounds):
        new_set = True
        for set in final_sets:
            if _in_set(set, lower) and not _in_set(set, upper):
                set[1] = upper
                new_set = False
                break
            elif not _in_set(set, lower) and _in_set(set, upper):  # noqa: RET508
                set[0] = lower
                new_set = False
                break
            elif _in_set(set, lower) and _in_set(set, upper):
                new_set = False
        if new_set:
            final_sets.append([lower, upper])
    return final_sets


def _in_set(set, val):
    if val >= set[0] and val <= set[1]:
        return True
    return None



from __future__ import annotations

import numpy as np

from ...core import OpticalGeometry
from ..orientation_techniques import OrientationTechniques
from .platform_pointing import PlatformPointing
from .platformlocator import PlatformLocation
from .positionandorientationarray import PositionAndOrientationArray
from .rotationmatrix import RotationMatrix


# -----------------------------------------------------------------------------
#           class Platform
# -----------------------------------------------------------------------------
class Platform:
    """
    The purpose of the Platform class is to capture the details of the physical platform carrying an instrument. Common
    examples of platforms used in atmospheric research are spacecraft, aircraft, balloons and ground-based sites. There
    are a myriad of details surrounding real platforms but the Platform class's only concern is to generate a list
    of instrument *time*, *location* and rotation matrices that properly orient :ref:`icf` for each simulated measurement.
    This list of platform state information is internally generated and then passed onto the next stage of analysis or retrieval.

    The general form of usage is to perform the following steps,

    #. Optional. Specify how an instrument is mounted on the platform. This defines the :ref:`icf`.
    #. Specify the universal times for a set of measurements
    #. Specify the position of the platform for a set of measurements using a variety of positioning techniques, see :ref:`positioning_technique` and :meth:`~.add_measurement_set`
    #. Specify the orientation of the platform for a set of measurements using a variety of orientation techniques, see :ref:`pointing_technique` and :meth:`~.add_measurement_set`.
    #. The orientations define the :ref:`pcf` and the positions define the :ref:`gcf`.
    #. Create the internal :class:`~.PositionAndOrientationArray` for the measurement set, see :meth:`~.make_observation_policy`.
    #. Convert the :class:`~.PositionAndOrientationArray` to arrays of position and instrument look vectors suitable for retrieval code.
    """

    # -----------------------------------------------------------------------------
    #           __init__
    # -----------------------------------------------------------------------------

    def __init__(
        self,
        observation_policy: PositionAndOrientationArray = None,
        platform_locator: PlatformLocation = None,
    ):
        """

        Parameters
        ----------
        platform_locator: PlatformLocation
            Default None. This optional parameter can be used to add a platform location object such as a satellite,
            aircraft or ground site.

        observation_policy: PositionAndOrientationArray
            Default None. This optional parameter can use an existing :ref:`observationpolicy_class` object rather than
            create a new empty instance.
        """

        # super().__init__()
        self._platform_pointing: PlatformPointing = PlatformPointing()
        self._platform_location: PlatformLocation = platform_locator
        self._orientationtechniques: OrientationTechniques = OrientationTechniques()
        self._position_and_orientation_array: PositionAndOrientationArray = (
            observation_policy
            if observation_policy is not None
            else PositionAndOrientationArray()
        )

    # ------------------------------------------------------------------------------
    #           platform_pointing
    # ------------------------------------------------------------------------------
    @property
    def platform_pointing(self) -> PlatformPointing:
        """
        Gets the internal :class:`~.PlatformPointing` object. This object manages all the rotation matrices used to
        transform between various frames.
        """
        return self._platform_pointing

    # ------------------------------------------------------------------------------
    #           platform_locator
    # ------------------------------------------------------------------------------
    @property
    def platform_locator(self) -> PlatformLocation:
        """
        Sets and gets the internal :class:`~.PlatformLocation` object. This field is set to None by default but can
        be set to a valid class if the user wishes to use a specialized class to get the platform position or orientation
        as a function of time.
        """
        return self._platform_location

    @platform_locator.setter
    def platform_locator(self, value):
        self._platform_location = value

    # ------------------------------------------------------------------------------
    #           observation_policy
    # ------------------------------------------------------------------------------
    @property
    def observation_policy(self) -> PositionAndOrientationArray:
        """
        Returns the current internal :ref:`observationpolicy_class` object.

        Returns
        -------
        PositionAndOrientationArray
            Returns the current internal :ref:`observationpolicy_class` object.
        """
        if (self._position_and_orientation_array is None) or (
            self._orientationtechniques._isdirty
        ):
            self.make_position_and_orientation_array()
        return self._position_and_orientation_array

    # ------------------------------------------------------------------------------
    #           platform_ecef_positions
    # ------------------------------------------------------------------------------
    @property
    def platform_ecef_positions(self) -> np.ndarray:
        """
        Returns the position of the platform for each RT calculation

        Returns
        -------
        np.ndarray(3,N)
            The X,Y,Z location of the platform for each RT calculation. The coordinates are in meters form the center of the Earth,
        """
        return self._position_and_orientation_array.ecef_positions()

    # ------------------------------------------------------------------------------
    #           icf_to_ecef_rotation_matrices
    # ------------------------------------------------------------------------------
    @property
    def icf_to_ecef_rotation_matrices(self) -> list[RotationMatrix]:
        """
        Returns the platform rotation matrices, one  matrix for each RT calculation

        Returns
        -------
        np.ndarray(3,N)
            The X,Y,Z icf_to_ecef_rotation_matrices of the platform.
        """
        return self._position_and_orientation_array.icf_to_ecef_rotation_matrix()

    # ------------------------------------------------------------------------------
    #           num_exposures
    # ------------------------------------------------------------------------------
    @property
    def numsamples(self) -> int:
        """
        Returns the number of samples/exposures in the current observation set

        Returns
        -------
        int
            The number of samples in the current observation set
        """
        return self.observation_policy.numsamples_in_observationset()

    # ------------------------------------------------------------------------------
    #           add_measurement_set
    # ------------------------------------------------------------------------------

    def add_measurement_set(
        self, utc, platform_position, platform_orientation, icf_orientation=None
    ):
        """
        Adds a set of *N* measurements definitions to the internal list of measurement sets. An overview of position and
        orientation techniques is given in :ref:`platforms_model`.

        Parameters
        ----------
        utc:Array, sequence or scalar of string, float, datetime or datetime64
            The set of universal times for the set of measurements. The array should be an array or sequence of any type that can be
            coerced by package sktime into a numpy array (N,) of datetime64. This includes strings, datetime, datetime64. Floating
            point values are also converted and are assumed to represent modified julian date. The number of universal time values
            defines the number of measurements in this set unless it is a scalar. In this case the time value is duplicated into an array with
            the same number of measurements as parameter *observer_positions*. All elements of the array should be of the same
            type and should represent universal time. Be wary of using datetime objects with explicit time-zones.
        observer_positions: Tuple[str, sequence]
            A two element tuple. The first element specifies the :ref:`positioning_technique` to be used to position the
            platform. The second element of the tuple is an array of arrays of float. The array of arrays can be any Python, list of lists, tuples or arrays
            that can be coerced into a two dimensional array of shape **(N, numparams)** where *N* is the number of
            measurements and must match the size of the *utc* parameter parameters and *numparams* is the number of parameters
            required by the chosen :ref:`positioning_technique`. The second element of the tuple can be dropped if the *positioning technique*
            is *from_platform*.
        platform_orientation:
            A three element tuple. The first element of the tuple is a string that specifies the :ref:`pointing_technique`. The second element
            of the tuple specifies :ref:`rollcontrol` and the third element is an array of arrays of float. The array of arrays can be any Python, list of lists, tuples or arrays
            that can be coerced into a two dimensional array of shape **(N, numparams)** where *N* is the number of
            measurements and must match the size of the *utc* parameter and *numparams* is the number of parameters
            required by the chosen :ref:`pointing_technique`
        instrument_internal_rotation:
            Optional. An array that specifies the internal rotation of the instrument within the :ref:`icf`. This is intended to
            provide support for tilting mirrors and turntables attached to the instrument that redirect the instrument boresight independently
            of the platform. The array specifies the azimuth, elevation and roll of the instrument boresight in the :ref:`icf`.
            The array is a sequence or array that can be sensibly coerced into an array of size (N,2) or (N,3) where N is the number of measurements.
            N can be 1 in which case the array size is broadcast to match the number of measurements inferred from the other parameters. Elements [:,0] is the azimuth in degrees  of the
            instrument boresight in the instrument control frame, left handed rotation around :math:`\\hat{z}_{ICF}`. Elements [:,1] are the elevation in
            degrees of the instrument boresight in the instrument control frame, left handed rotation around the rotated :math:`\\hat{y}_{ICF}` axis.
            Elements[:,2], which are are optional, are the roll of the instrument boresight in degrees, right handed rotation around the rotated  :math:`\\hat{x}_{ICF}` axis.
            The roll defaults to 0.0 if not supplied.

        """
        self._orientationtechniques.add_measurement_set(
            utc,
            platform_position,
            platform_orientation,
            icf_orientation=icf_orientation,
        )

    # ------------------------------------------------------------------------------
    #           make_observation_policy
    # ------------------------------------------------------------------------------

    def make_position_and_orientation_array(self) -> PositionAndOrientationArray:
        """
        Takes all the measurements from previous calls to :meth:`~.Platform.add_measurement_set` and converts them into
        a list of measurements stored inside the internal instance of :class:`~.PositionAndOrientationArray` object. This list of
        measurements can be converted to other formats, such as **OpticalGeometry** required by other parts of the
        ``skretrieval`` package. This method clears all the measurements that have been previously added.
        """

        self._position_and_orientation_array.clear()
        self._orientationtechniques.make_observation_set(self)
        return self._position_and_orientation_array

    # ------------------------------------------------------------------------------
    #           make_optical_geometry
    # ------------------------------------------------------------------------------
    def make_optical_geometry(self) -> list[OpticalGeometry]:
        """
        Takes all the measurements from previous calls to :meth:`~.Platform.add_measurement_set` and returns them as
        a list of OpticalGeometry values which can be used by various retrievals.  The internal :class:`~.PositionAndOrientationArray` is
        also created at this time using method :meth:`~Platform.make_observation_policy`.
        """

        position_and_orientation_array = self.make_position_and_orientation_array()
        return position_and_orientation_array.to_optical_geometry()

    # -----------------------------------------------------------------------------
    #               make_ecef_position_and_lookvector
    # -----------------------------------------------------------------------------
    def make_ecef_position_and_lookvector(
        self, icf_lookvectors=None, one_to_one: bool = False
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        position_and_orientation_array = self.make_position_and_orientation_array()
        return position_and_orientation_array.to_ecef_position_and_look(
            icf_lookvectors=icf_lookvectors, one_to_one=one_to_one
        )

    # -----------------------------------------------------------------------------
    #               def add_current_state(self  )->int:
    # -----------------------------------------------------------------------------
    def add_current_platform_state(self) -> int:
        """
        Fetches the current platform position and orientation and uses this as a new sample in the observation
        policy.

        Returns
        -------
        int
            The number of samples in the current observation policy

        """
        return self._position_and_orientation_array.add_current_state(
            self._platform_pointing
        )

    # -----------------------------------------------------------------------------
    #               clear_states(self):
    # -----------------------------------------------------------------------------
    def clear_states(self):
        """
        Clears all of the internally cached measurement states. This should be called when
        re-using a platform object to create a new measurement set.
        """
        self._orientationtechniques.clear()
        self._position_and_orientation_array.clear()

    # ------------------------------------------------------------------------------
    #           icf_to_ecef
    # ------------------------------------------------------------------------------
    def icf_to_ecef(self, los_icf: np.ndarray) -> np.ndarray:
        """
        Returns the lines of sight in geographic geocentric ECEF coordinates of the lines of sight specified
        in the instrument control frame.

        Parameters
        ----------
        los_icf : np.ndarray( 3,Nlos)
            A 2-D array of N unit vectors expressed in the instrument control frame

        Returns
        -------
        np.ndarray (3,Nlos, Ntime)
            A numpy array is returned with the 3 element geographic, geocentric, line of sight unit vector for each
            input lines of sight and each time in the current observation set.
        """
        Nlos = los_icf.shape[1]
        obs = (
            self.observation_policy
        )  # Get the observation set used for this back_end_radiance calculation                                                                                    # Get the number of instantaneous lines of sight from the second dimension of the array. LOS are specified in the instrument control frame
        Nt = (
            obs.numsamples_in_observationset()
        )  # Get the number of exposures/samples in the observation set
        rot = (
            obs.icf_to_ecef_rotation_matrix()
        )  # Copy the rotation matrices for each sample in the observation set.
        all_los = np.full(
            (3, Nlos, Nt), np.nan
        )  # Create an array to hold the lines of sight for all instantaneous directions and all times
        for i in range(Nt):  # for each time
            R = rot[
                i
            ]  # Get the platform rotation matrix that converts ICF vectors to ECEF vectors
            instantaneouslos = (
                R.R @ los_icf
            )  # Convert all the instrument control frame instantaneous lines of sight for this sample to geocentric geographic lines of sight
            all_los[
                :, :, i
            ] = instantaneouslos  # and save the geo unit vectors in the array.
        return all_los



from __future__ import annotations

import math
import numbers
from datetime import datetime
from typing import Any, Sequence

import jdcal
import numpy as np

from .datetime64 import datetime64_to_datetime


# ------------------------------------------------------------------------------
#           datetime64_to_jd
# ------------------------------------------------------------------------------
def datetime_to_jd(t: datetime) -> tuple[float, float]:
    """
    Converts a scalar datetime.datetime object to julian date which is returned as
    a two element floating point tuple. No attempt has been made to manage timezones
    inside the datetime object. We recommend that you don't use objects with time-zones in this code.

    Parameters
    ----------
    ut1 : datetime.datetime
        The universal time to be converted to julian date. The Universal time is assumed to follow the
        Gregorian calendar

    Returns
    -------
    Tuple( float, float)
        Returns a two element floating point tuple. The first element contains the high order (julian days) part of
        the julian date. The second element contains the low order fraction of a day.
    """

    jd1, jd2 = jdcal.gcal2jd(t.year, t.month, t.day)
    jd1 = jd1 + jd2 - 0.5
    frac = (
        t.hour * 3600.0 + t.minute * 60.0 + t.second + t.microsecond / 1.0e6
    ) / 86400.0 + 0.5
    if frac >= 1.0:
        jd1 += 1.0
        frac -= 1.0
    return jd1, frac


# ------------------------------------------------------------------------------
#           datetime64_to_jd
# ------------------------------------------------------------------------------
def datetime64_to_jd(ut1: np.datetime64) -> tuple[float, float]:
    """
    Converts a scalar numpy datetime64 object to julian date which is returned as
    a two element floating point tuple. No attempt has been made to manage timezones
    inside the datetime64 object. We recommend you don't use objects with time-zones

    Parameters
    ----------
    ut1 : numpy.datetime64
        The universal time (strictly UT1) to be converted to julian date.

    Returns
    -------
    Tuple( float, float)
        Returns a two element floating point tuple. The first element contains the high order (julian days) part of
        the julian date. The second element contains the low order fraction of a day.
    """
    t = datetime64_to_datetime(ut1)
    return datetime_to_jd(t)


# ------------------------------------------------------------------------------
#           isoformat_to_jd
# ------------------------------------------------------------------------------
def isoformat_to_jd(utc: str) -> tuple[float, float]:
    """
    Converts a string representing a UT instant encoded in isoformat to a julian date which is returned as
    a two element floating point tuple. No attempt has been made to manage timezones
    inside the datetime object. We recommend that you don't use objects with time-zones in this code.

    Parameters
    ----------
    ut1 : str
        The universal time (strictly UT1) to be converted to julian date. The string is encoded is isoformat,
        eg '2009-11-15T15:36:49.000123'

    Returns
    -------
    Tuple( float, float)
        Returns a two element floating point tuple. The first element contains the high order (julian days) part of
        the julian date. The second element contains the low order fraction of a day.
    """

    t = datetime.fromisoformat(str(utc))  # convert the string to datetime
    return datetime_to_jd(t)  # and from datetime to mjd


# ------------------------------------------------------------------------------
#           number_to_jd
# ------------------------------------------------------------------------------
def number_to_jd(utc: float) -> tuple[float, float]:
    """
    Converts a number to a julian date which is returned as a two element floating point tuple. Numbers less than
    500,000 are assumed to represent modified julian date. Numbers greater than 500,000 are assumed to be julian dates.
    The calculation is useful for converting floating point values to the two element tuples used by the novas
    and sofa software.

    Parameters
    ----------
    ut1 : float
        The universal time to be converted to julian date.

    Returns
    -------
    Tuple( float, float)
        Returns a two element floating point tuple. The first element contains the high order (julian days) part of
        the julian date. The second element contains the low order fraction of a day.
    """

    x = float(utc)
    jd1 = float(math.floor(x))
    jd2 = x - jd1
    if jd1 < 500000.0:
        jd1 += 2400000.0
        jd2 += 0.5
        if jd2 >= 1.0:
            jd1 += 1.0
            jd2 -= 1.0
    return jd1, jd2  # and from datetime to mjd


# ------------------------------------------------------------------------------
#           _iterate_over_utc_array
# ------------------------------------------------------------------------------
def _iterate_over_utc_array(utc, elementfunc):
    s = utc.shape  # get the shape
    jd1 = np.zeros(
        s, dtype=np.float64
    )  # create the output array for the high order julian date components
    jd2 = np.zeros(
        s, dtype=np.float64
    )  # create the output array for the low order julian date component
    with np.nditer(
        [utc, jd1, jd2],
        flags=["refs_ok"],
        op_flags=[["readonly"], ["writeonly"], ["writeonly"]],
    ) as it:  # create the numpy iterator object
        for a, b, c in it:  # and convert each elemnet
            j1, j2 = elementfunc(
                a.item()
            )  # from the numpy array element datetime to jd
            b[...] = j1
            c[...] = j2
        jd1 = it.operands[1]
        jd2 = it.operands[2]
    return jd1, jd2


# ------------------------------------------------------------------------------
#           utc_to_jd
# ------------------------------------------------------------------------------
def ut_to_jd(ut: np.ndarray | (list[Any] | Any)):
    """
    A convenience function that converts various arrays or scalar representations of UT to julian date. Each Julian date is returned
    as two floating point components, `jd1` and `jd2`. If the input is an array of value sthen jd1 and jd2 will be returned as
    floating point arrays with the same shape and size as the input array. If the inputs are scalar values then `jd1` and `jd2` will
    be returned as scalar floating point values.

    Parameters
    ----------
    utc : scalar, list, tuple or array
        The input time which represents a universal time (strictly UT1). It can be represented by
        (i) a string in a supported python datetime.isoformat
        (ii) a number which is assumed to represent a julian date if it is greater than 500,000 or a modified julian date if it less than 500,000,
        (iii) a datetime.datetime object representing UT
        (iv) a numpy.datetime64 object representing UT

        The *ut* object can be a scalar, list, tuple or numpy array. The same representation format must be used for
        all objects in the arrays and sequences.

    Returns
    -------
    scalar or numpy.ndarray of float
        The mjd is returned as a scalar if the input was a scalar or as a numpy array for input sequences and arrays
    """

    if (
        type(ut) is np.ndarray
    ):  # Handle an numpy array of values, datetime64, datetime objects or mjd numbers
        if ut.dtype.kind == "M":  # if we have a numpy array of datetime64 object
            jd1, jd2 = _iterate_over_utc_array(
                ut, datetime64_to_jd
            )  # Then convert all the elements to julian date
        elif (
            (ut.dtype.kind == "O") and (len(ut) > 0) and (type(ut[0]) is datetime)
        ):  # otherwise if we have a numpy array of  datetime objects
            jd1, jd2 = _iterate_over_utc_array(
                ut, datetime_to_jd
            )  # create the output array
        elif ut.dtype.kind == "U":
            jd1, jd2 = _iterate_over_utc_array(
                ut, isoformat_to_jd
            )  # otherwise if the array is strings
        elif (ut.dtype.kind == "f") or (ut.dtype.kind == "i"):
            jd1, jd2 = _iterate_over_utc_array(
                ut, number_to_jd
            )  # otherwise if the array is strings
    elif isinstance(ut, str):  # if we have a single string
        jd1, jd2 = isoformat_to_jd(ut)  # then convert the string to datetime
    elif isinstance(ut, Sequence):  # if we have a list or tuple but not numpy array
        newutc = np.array(ut)  # then convert the array to numpy array
        jd1, jd2 = ut_to_jd(newutc)  # and convert the numpy array to mjd
    else:  # Its not a numpy array, its not a sequence, so assume its a scalar
        if type(ut) is datetime:  # if its a datetime object
            jd1, jd2 = datetime_to_jd(ut)  # to mjd as a floating point
        elif type(ut) is np.datetime64:  # if we have a single numpy.datetime64
            jd1, jd2 = datetime64_to_jd(ut)  # then convert it
        elif isinstance(ut, numbers.Number):  # if we have a number
            jd1, jd2 = number_to_jd(ut)
        else:  # otherwise
            jd1 = math.nan
            jd2 = math.nan
            msg = "utc_to_jd, unsupported data type {}. Cannot convert to julian date".format(
                str(type(ut))
            )
            raise Exception(msg)
    return jd1, jd2



from __future__ import annotations

from datetime import datetime
from math import radians

import numpy as np

import skretrieval.time

from .satellitebase import SatelliteBase
from .satellitekepler import SatelliteKepler
from .satellitesgp4 import SatelliteSGP4


# ------------------------------------------------------------------------------
#           class SatelliteMolniya
# ------------------------------------------------------------------------------
class SatelliteMolniya(SatelliteBase):
    """
    Implements a Molniya orbit. This model can internally utilize either a Keplerian or SGP4 orbit propagator.
    The SGP4 propagator will capture gravitational perturbations on the orbit but the Kepler will not. The user will
    typically initialize the molniya orbit in the constructor but this simply calls
    :meth:`~.set_molniya_from_elements`. A full desccription of the optional keyword arguments
    is given in that function. A sensible default Molniya orbit is provided in the defaults
    """

    def __init__(
        self,
        utc: datetime | (np.datetime64 | (float | str)),
        orbittype="kepler",
        period_from_seconds: float = 718.0 * 60.0,
        inclination_user_defined: float = radians(63.4),
        argument_of_perigee: float = radians(270.0),
        eccentricity: float = 0.74,
        sgp4_bstar_drag: float = 0.0,
        **kwargs,
    ):
        super().__init__()
        ut = skretrieval.time.ut_to_datetime(utc)
        self.satellite = None
        self.set_molniya_from_elements(
            ut,
            orbittype=orbittype,
            period_from_seconds=period_from_seconds,
            inclination_user_defined=inclination_user_defined,
            argument_of_perigee=argument_of_perigee,
            eccentricity=eccentricity,
            sgp4_bstar_drag=sgp4_bstar_drag,
            **kwargs,
        )

    # -----------------------------------------------------------------------------
    #           update_eci_position
    # -----------------------------------------------------------------------------

    def update_eci_position(self, tnow: datetime):
        """
        Updates the eciposition and ecivelocity of the satellite at the given instant in time.

        Parameters
        ----------
        tnow : datetime
            Updates the ECI eciposition and ECI ecivelocity of the stellite to this time
        """
        self.satellite.update_eci_position(tnow)
        self._set_current_state(
            self.satellite.eciposition(), self.satellite.ecivelocity(), tnow
        )

    # ------------------------------------------------------------------------------
    #           period
    # ------------------------------------------------------------------------------

    def period(self) -> float:
        """
        Return the period of the satellite.  For low earth orbit spacecraft
        this is accurate to about 0.1 seconds.

        Returns
        -------
        datetime.timedelta
            The period of the satellite
        """
        return self.satellite.period()

    # ------------------------------------------------------------------------------
    #           eccentricity
    # ------------------------------------------------------------------------------

    def eccentricity(self) -> float:
        return self.satellite.eccentricity()

    # -----------------------------------------------------------------------------
    #           set_sun_sync_from_elements
    # -----------------------------------------------------------------------------

    def set_molniya_from_elements(
        self,
        utc: datetime,
        orbittype="kepler",
        period_from_seconds: float = 718.0 * 60.0,
        inclination_user_defined: float = radians(63.4),
        argument_of_perigee: float = radians(270.0),
        eccentricity: float = 0.74,
        sgp4_bstar_drag: float = 0.0,
        **kwargs,
    ):
        """
        Defines the Molniya orbit using Keplerian orbital elements. A default, standard Molniya orbit is provided via the default
        although the user must supply the right ascension of ascending node using one of the standard :class:`~.SatelliteKepler`
        options (i) **localtime_of_ascending_node_hours**, (ii) **longitude_of_ascending_node_degrees** or (iii) **right_ascension_ascending_node**

        The Molniya orbit is always kick-started using Keplerian orbital elements. It can continue to use the keplerian
        orbit to propagate the position and velocity or it can switch to an SGP4 model orbit propagator.

        Parameters
        ----------
        utc : datetime.datetime
            The universal time of the orbital elements.
        orbittype: str
            Specifies the type of orbit predictor to be used to propagate the orbit. It can be either 'kepler' or 'SGP4'. Default is 'kepler'
        sgp4_bstar_drag: float
            Optional argument that is only used if the `sgp4` propagator is selected. This value is used as the bstar drag
            term in the SGP4 orbit predictor. The default value is 0.0 implying drag is ignored.
        kwargs: extra key word arguments.
            These keywords are from the keyword options in :meth:`.SatelliteKepler.from_elements`.
        """
        kwargs["period_from_seconds"] = period_from_seconds
        kwargs["inclination_radians"] = inclination_user_defined
        kwargs["inclination_is_sun_sync"] = False
        kwargs["argument_of_perigee"] = argument_of_perigee
        kwargs["eccentricity"] = eccentricity
        kepler = SatelliteKepler(utc, **kwargs)
        if orbittype.lower() == "kepler":
            self.satellite = kepler
        elif orbittype.lower() == "sgp4":
            sgp4 = SatelliteSGP4()
            sgp4.from_kepler_orbit(kepler, bstar=sgp4_bstar_drag)
            self.satellite = sgp4
        else:
            msg = f"the requested type of orbit predictor {orbittype} is not supported"
            raise Exception(msg)



from __future__ import annotations

import numpy as np
import sasktran as sk
import sasktran2 as sk2

from skretrieval.geodetic import geodetic


def _limb_viewing_los(
    los: sk.LineOfSight, forced_sun: np.array = None
) -> sk2.TangentAltitudeSolar:
    tangent_location = los.tangent_location()

    # TODO: Get this from astropy
    sun = None if forced_sun is None else forced_sun

    cos_sza = np.dot(tangent_location.local_up, sun)

    los_projected = los.look_vector - tangent_location.local_up * (
        los.look_vector.dot(tangent_location.local_up)
    )
    los_projected /= np.linalg.norm(los_projected)

    sun_projected = sun - tangent_location.local_up * (
        sun.dot(tangent_location.local_up)
    )
    sun_projected /= np.linalg.norm(sun_projected)

    y_axis = np.cross(tangent_location.local_up, sun_projected)

    saa = np.arctan2(y_axis.dot(los_projected), sun_projected.dot(los_projected))

    obs_geo = geodetic()
    obs_geo.from_xyz(los.observer)

    return sk2.TangentAltitudeSolar(
        tangent_altitude_m=tangent_location.altitude,
        relative_azimuth=saa,
        observer_altitude_m=obs_geo.altitude,
        cos_sza=cos_sza,
    )


def _ground_viewing_los(
    los: sk.LineOfSight, forced_sun: np.array = None
) -> sk2.GroundViewingSolar:
    ground_location = los.ground_intersection()

    # TODO: Get this from astropy
    sun = None if forced_sun is None else forced_sun

    cos_sza = np.dot(ground_location.local_up, sun)

    los_projected = los.look_vector - ground_location.local_up * (
        los.look_vector.dot(ground_location.local_up)
    )
    los_projected /= np.linalg.norm(los_projected)

    sun_projected = sun - ground_location.local_up * (sun.dot(ground_location.local_up))
    sun_projected /= np.linalg.norm(sun_projected)

    y_axis = np.cross(ground_location.local_up, sun_projected)

    saa = np.arctan2(y_axis.dot(los_projected), sun_projected.dot(los_projected))

    obs_geo = geodetic()
    obs_geo.from_xyz(los.observer)

    cos_zenith = np.abs(np.dot(los.look_vector, ground_location.local_up))

    return sk2.GroundViewingSolar(
        cos_sza,
        saa,
        cos_zenith,
        np.linalg.norm(obs_geo.location) - np.linalg.norm(ground_location.location),
    )


def convert_sasktran_legacy_geometry(
    legacy_geometry: sk.Geometry,
) -> sk2.ViewingGeometry:
    """

    Parameters
    ----------
    legacy_geometry : sk.Geometry
        _description_

    Returns
    -------
    sk2.ViewingGeometry
        _description_
    """
    sk2_geometry = sk2.ViewingGeometry()

    for los in legacy_geometry.lines_of_sight:
        if los.tangent_location() is not None:
            ray = _limb_viewing_los(los, legacy_geometry.sun)
        else:
            ray = _ground_viewing_los(los, legacy_geometry.sun)

        sk2_geometry.add_ray(ray)

    return sk2_geometry



from __future__ import annotations

from .platform import Platform
from .platform_pointing import PlatformPointing
from .platformlocator import PlatformLocation
from .rotationmatrix import RotationMatrix, UnitVectors



from __future__ import annotations

import numpy as np
import sasktran as sk

from skretrieval.legacy.util import convert_sasktran_legacy_geometry


def test_convert_sasktran_legacy_geometry():
    tanalts_km = np.arange(10, 60, 1.0)

    geo = sk.VerticalImage()

    geo.from_sza_saa(
        sza=60,
        saa=60,
        lat=0,
        lon=0,
        tanalts_km=tanalts_km,
        mjd=54372,
        locallook=0,
        satalt_km=600,
        refalt_km=20,
    )

    _ = convert_sasktran_legacy_geometry(geo)



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry
from skretrieval.core.sasktranformat import SASKTRANRadiance


class Sensor(ABC):
    """
    A sensor is an object which takes in radiances at the front of an aperture, and returns backend radiances
    observed by the sensor.
    """

    @abstractmethod
    def model_radiance(
        self,
        radiance: SASKTRANRadiance,
        orientation: OpticalGeometry,
    ) -> radianceformat.RadianceBase:
        """
        Takes in high resolution radiances at the front of the aperture and converts them to what is observed by the
        sensor.

        Parameters
        ----------
        radiance : SasktranFormat
            Input radiances in the format of the SASKTRAN radiative transfer model

        orientation : OpticalGeometry
            Orientation of the sensor

        Returns
        -------
        radianceformat.RadianceBase
            Output L1 radiances in a format specific to the sensor.  The format is defined by `radiance_format`

        """

    @staticmethod
    @abstractmethod
    def radiance_format():
        """
        Returns the format that model_radiance returns data in

        Returns
        -------
        A specialized instance of radianceformat.RadianceBase
        """
